######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains model for export preview tables."""
from itertools import takewhile
from operator import methodcaller
from PySide6.QtCore import QAbstractItemModel, QModelIndex, Qt


class PreviewTreeModel(QAbstractItemModel):
    TABLE_ROLE = Qt.ItemDataRole.UserRole + 1

    def __init__(self):
        super().__init__()
        self._tables = {}
        self._table_names = {}
        self._mapping_names = []

    def add_or_update_tables(self, mapping_name, tables):
        """
        Adds new tables to the model.

        Args:
            mapping_name (str): mapping's name
            tables (dict): mapping from table name to table data
        """
        if None in tables:
            tables["<anonymous table>"] = tables.pop(None)
        if mapping_name not in self._tables:
            new_names = list(self._mapping_names)
            new_names.append(mapping_name)
            new_names.sort(key=methodcaller("lower"))
            row = len(list(takewhile(lambda x: x[0] == x[1], zip(new_names, self._mapping_names))))
            self.beginInsertRows(QModelIndex(), row, row)
            self._tables[mapping_name] = {}
            self._mapping_names = new_names
            self._table_names[mapping_name] = []
            self.endInsertRows()
        parent_index = self.index(self._mapping_names.index(mapping_name), 0)
        for old_name in list(self._tables[mapping_name]):
            if old_name not in tables:
                names = self._table_names[mapping_name]
                row = names.index(old_name)
                self.beginRemoveRows(parent_index, row, row)
                del self._tables[mapping_name][old_name]
                self._table_names[mapping_name] = names[:row] + names[row + 1 :]
                self.endRemoveRows()
        for name, table in tables.items():
            if name in self._tables[mapping_name]:
                self._tables[mapping_name][name] = table
                row = self._table_names[mapping_name].index(name)
                index = self.index(row, 0, parent_index)
                self.dataChanged.emit(index, index, [self.TABLE_ROLE])
            else:
                self._tables[mapping_name][name] = table
                new_names = list(self._table_names[mapping_name])
                new_names.append(name)
                new_names.sort(key=methodcaller("lower"))
                row = len(list(takewhile(lambda x: x[0] == x[1], zip(new_names, self._table_names[mapping_name]))))
                self.beginInsertRows(parent_index, row, row)
                self._table_names[mapping_name] = new_names
                self.endInsertRows()

    def columnCount(self, parent=QModelIndex()):
        return 1

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            mapping_name_row = index.internalId() - 1
            if mapping_name_row < 0:
                return self._mapping_names[index.row()]
            mapping_name = self._mapping_names[mapping_name_row]
            return self._table_names[mapping_name][index.row()]
        if role == self.TABLE_ROLE and index.internalId() > 0:
            mapping_name = self._mapping_names[index.internalId() - 1]
            return self._tables[mapping_name][self._table_names[mapping_name][index.row()]]
        return None

    def clear(self):
        """
        Clears the model.
        """
        self.beginResetModel()
        self._tables = {}
        self._table_names = {}
        self._mapping_names = []
        self.endResetModel()

    def flags(self, parent=QModelIndex()):
        if parent.internalId() > 0:
            return super().flags(parent) | Qt.ItemNeverHasChildren
        return super().flags(parent)

    def has_name(self, name):
        """Returns True if the model has a mapping with given name.

        Args:
            name (str): mapping's name to look for

        Returns:
            bool: True if mapping was found, False otherwise
        """
        return name in self._mapping_names

    def index(self, row, column, parent=QModelIndex()):
        if not parent.isValid():
            return self.createIndex(row, column, 0)
        if parent.internalId() > 0:
            return QModelIndex()
        return self.createIndex(row, column, parent.row() + 1)

    def parent(self, index):
        if index.internalId() == 0:
            return QModelIndex()
        row = index.internalId() - 1
        return self.createIndex(row, 0, 0)

    def rename_mappings(self, new_names):
        """
        Renames a mapping.

        Args:
            new_names (list of str): mapping names; only a single name is expected to be actually new

        Returns:
            tuple of str: old name, new name
        """
        new_names = sorted(new_names, key=methodcaller("lower"))
        olds = set(self._mapping_names)
        news = set(new_names)
        try:
            old_name = next(iter(olds - news))
        except StopIteration:
            # A disabled mapping was renamed.
            return "", ""
        new_name = next(iter(news - olds))
        old_row = self._mapping_names.index(old_name)
        self.beginRemoveRows(QModelIndex(), old_row, old_row)
        self._mapping_names = self._mapping_names[:old_row] + self._mapping_names[old_row + 1 :]
        tables = self._tables.pop(old_name)
        table_names = self._table_names.pop(old_name)
        self.endRemoveRows()
        new_row = new_names.index(new_name)
        self.beginInsertRows(QModelIndex(), new_row, new_row)
        self._mapping_names = new_names
        self._tables[new_name] = tables
        self._table_names[new_name] = table_names
        self.endInsertRows()
        return old_name, new_name

    def remove_mapping(self, name):
        """
        Deletes a mapping from the model.

        Args:
            name (str): mapping's name
        """
        try:
            mapping_name_row = self._mapping_names.index(name)
        except ValueError:
            # Ignoring. Mappings are being removed while the previews are still loading.
            return
        parent_index = self.index(mapping_name_row, 0)
        self.beginRemoveRows(parent_index, 0, len(self._tables[name]) - 1)
        self._tables[name].clear()
        self._table_names[name].clear()
        self.endRemoveRows()
        self.beginRemoveRows(QModelIndex(), mapping_name_row, mapping_name_row)
        del self._tables[name]
        del self._table_names[name]
        del self._mapping_names[mapping_name_row]
        self.endRemoveRows()

    def rowCount(self, parent=QModelIndex()):
        if not parent.isValid():
            return len(self._mapping_names)
        mapping_name = self._mapping_names[parent.row()]
        return len(self._table_names[mapping_name])



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`SpecificationEditorWindow`."""
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QFileDialog, QHeaderView
from spinetoolbox.helpers import disconnect
from spinetoolbox.project_item.specification_editor_window import (
    ChangeSpecPropertyCommand,
    SpecificationEditorWindowBase,
)
from ..data_transformer_specification import DataTransformerSpecification
from ..settings import EntityClassRenamingSettings, ParameterRenamingSettings, ValueTransformSettings
from .class_rename import ClassRename
from .parameter_rename import ParameterRename
from .value_transformation import ValueTransformation

_FILTER_NAMES = ("Rename entity classes", "Rename parameters", "Transform values")

_SETTINGS_CLASSES = dict(
    zip(_FILTER_NAMES, (EntityClassRenamingSettings, ParameterRenamingSettings, ValueTransformSettings))
)

_CLASSES_TO_DISPLAY_NAMES = {class_: name for name, class_ in _SETTINGS_CLASSES.items()}


class SpecificationEditorWindow(SpecificationEditorWindowBase):
    """Data transformer's specification editor."""

    def __init__(self, toolbox, specification=None, item=None, urls=None):
        """
        Args:
            toolbox (ToolboxUI): Toolbox main window
            specification (ProjectItemSpecification, optional): transformer specification
            item (ProjectItem, optional): invoking project item, if window was opened from its properties tab
            urls (dict, optional): a mapping from provider name to database URL
        """

        super().__init__(toolbox, specification, item)
        self._settings = {name: None for name in _FILTER_NAMES}
        self._changing_docks = [
            self._ui.load_database_dock,
            self._ui.possible_parameters_dock,
            self._ui.parameter_rename_dock,
            self._ui.value_transformation_dock,
            self._ui.value_instructions_dock,
            self._ui.possible_classes_dock,
            self._ui.class_rename_dock,
        ]
        for dock in self._changing_docks:
            dock.hide()
        self._urls = {}
        self._filter_sub_interfaces = {}
        self._current_filter_name = None
        self.takeCentralWidget().deleteLater()
        self._ui.filter_combo_box.addItems(_FILTER_NAMES)
        if specification is not None and specification.settings is not None:
            settings = specification.settings
            filter_name = _CLASSES_TO_DISPLAY_NAMES[type(settings)]
            self._settings[filter_name] = settings
        else:
            filter_name = _FILTER_NAMES[0]
        self._ui.load_url_from_fs_button.clicked.connect(self._load_url_from_filesystem)
        self._ui.load_data_button.clicked.connect(self._load_data)
        self._ui.database_url_combo_box.model().rowsInserted.connect(
            lambda *args: self._ui.load_data_button.setEnabled(True)
        )
        self._ui.database_url_combo_box.addItems(urls if urls is not None else [])
        self._ui.filter_combo_box.currentTextChanged.connect(self._change_filter_widget)
        self._set_current_filter_name(filter_name)
        for view in (
            self._ui.available_parameters_tree_view,
            self._ui.available_classes_tree_widget,
        ):
            view.header().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        for view in (
            self._ui.parameter_rename_table_view,
            self._ui.transformations_table_view,
            self._ui.class_rename_table_view,
        ):
            view.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)

    @property
    def settings_group(self):
        return "dataTransformerSpecificationWindow"

    def _make_ui(self):
        from ..ui.specification_editor_widget import Ui_MainWindow  # pylint: disable=import-outside-toplevel

        return Ui_MainWindow()

    def _make_new_specification(self, spec_name):
        """See base class."""
        description = self._spec_toolbar.description()
        filter_name = self._ui.filter_combo_box.currentText()
        if not filter_name:
            filter_settings = None
        else:
            interface = self._filter_sub_interfaces[filter_name]
            filter_settings = interface.settings()
        return DataTransformerSpecification(spec_name, filter_settings, description)

    @Slot(str)
    def _change_filter_widget(self, filter_name):
        """
        Pushes a command to change the current interface to undo stack.

        Args:
            filter_name (str): widget's filter name
        """
        if self._current_filter_name == filter_name:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_current_filter_name, filter_name, self._current_filter_name, "change filter"
            )
        )

    def _set_current_filter_name(self, filter_name):
        """Shows/hides current dock widgets depending on selected filter.

        Args:
            filter_name (str): filter's name
        """
        previous_interface = self._filter_sub_interfaces.get(self._current_filter_name)
        if previous_interface is not None:
            previous_interface.tear_down()
        self._current_filter_name = filter_name
        with disconnect(self._ui.filter_combo_box.currentTextChanged, self._change_filter_widget):
            self._ui.filter_combo_box.setCurrentText(filter_name)
        interface = self._filter_sub_interfaces.get(filter_name)
        if interface is None:
            interface = dict(zip(_FILTER_NAMES, (ClassRename, ParameterRename, ValueTransformation)))[filter_name](
                self._ui, self._undo_stack, self._settings[filter_name], self
            )
            self._filter_sub_interfaces[filter_name] = interface
        interface.show()

    @Slot(bool)
    def _load_url_from_filesystem(self, _):
        path = self._browse_database()
        if not path:
            return
        url = "sqlite:///" + path
        self._ui.database_url_combo_box.addItem(url)
        self._urls[url] = url

    def _browse_database(self):
        """
        Queries a database file from the user.

        Returns:
            str: path to database file
        """
        initial_path = self._toolbox.project().project_dir
        return QFileDialog.getOpenFileName(self, "Select database", initial_path, "sqlite (*.sqlite)")[0]

    @Slot(bool)
    def _load_data(self, _):
        """Sends selected database URL to current filter widget so it can load relevant data."""
        filter_name = self._ui.filter_combo_box.currentText()
        interface = self._filter_sub_interfaces[filter_name]
        interface.load_data(self._ui.database_url_combo_box.currentText())

    def _restore_dock_widgets(self):
        docks = {
            Qt.LeftDockWidgetArea: (
                self._ui.type_dock,
                self._ui.possible_classes_dock,
                self._ui.possible_parameters_dock,
                self._ui.load_database_dock,
            ),
            Qt.RightDockWidgetArea: (
                self._ui.parameter_rename_dock,
                self._ui.class_rename_dock,
                self._ui.value_transformation_dock,
                self._ui.value_instructions_dock,
            ),
        }
        for area, area_docks in docks.items():
            for dock in area_docks:
                self.addDockWidget(area, dock)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for View project item."""
import os
from tempfile import TemporaryDirectory
import unittest
from unittest.mock import MagicMock, NonCallableMagicMock
from PySide6.QtWidgets import QApplication
import spine_items.resources_icons_rc  # pylint: disable=unused-import
from spine_items.view.item_info import ItemInfo
from spine_items.view.view import View
from spine_items.view.view_factory import ViewFactory
from ..mock_helpers import create_mock_project, create_mock_toolbox, mock_finish_project_item_construction


class TestView(unittest.TestCase):
    def setUp(self):
        """Set up."""
        self.toolbox = create_mock_toolbox()
        factory = ViewFactory()
        item_dict = {"type": "View", "description": "", "x": 0, "y": 0}
        self._temp_dir = TemporaryDirectory()
        self.project = create_mock_project(self._temp_dir.name)
        self.toolbox.project.return_value = self.project
        self.view = factory.make_item("V", item_dict, self.toolbox, self.project)
        self._properties_widget = mock_finish_project_item_construction(factory, self.view, self.toolbox)

    def tearDown(self):
        self._temp_dir.cleanup()

    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_item_type(self):
        self.assertEqual(View.item_type(), ItemInfo.item_type())

    def test_item_dict(self):
        """Tests Item dictionary creation."""
        d = self.view.item_dict()
        a = ["type", "description", "x", "y"]
        for k in a:
            self.assertTrue(k in d, f"Key '{k}' not in dict {d}")

    def test_notify_destination(self):
        self.view.logger.msg = MagicMock()
        self.view.logger.msg_warning = MagicMock()
        source_item = NonCallableMagicMock()
        source_item.name = "source name"
        source_item.item_type = MagicMock(return_value="Data Connection")
        self.view.notify_destination(source_item)
        self.view.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>Data Connection</b> and"
            " a <b>View</b> has not been implemented yet."
        )
        source_item.item_type = MagicMock(return_value="Importer")
        self.view.notify_destination(source_item)
        self.view.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>Importer</b> and a <b>View</b> has not been implemented yet."
        )
        source_item.item_type = MagicMock(return_value="Data Store")
        self.view.notify_destination(source_item)
        self.view.logger.msg.emit.assert_called_with(
            "Link established. You can visualize Data Store <b>source name</b> in View <b>V</b>."
        )
        source_item.item_type = MagicMock(return_value="Tool")
        self.view.notify_destination(source_item)
        self.view.logger.msg.emit.assert_called_with(
            "Link established. You can visualize the output from Tool <b>source name</b> in View <b>V</b>."
        )

    def test_rename(self):
        """Tests renaming a View."""
        self.view.activate()
        expected_name = "ABC"
        expected_short_name = "abc"
        self.view.rename(expected_name, "")
        # Check name
        self.assertEqual(expected_name, self.view.name)  # item name
        self.assertEqual(expected_name, self.view.get_icon().name())  # name item on Design View
        # Check data_dir
        expected_data_dir = os.path.join(self.project.items_dir, expected_short_name)
        self.assertEqual(expected_data_dir, self.view.data_dir)  # Check data dir


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'specification_editor_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QComboBox, QDockWidget,
    QFormLayout, QHBoxLayout, QHeaderView, QLabel,
    QListWidget, QListWidgetItem, QMainWindow, QPushButton,
    QSizePolicy, QSpacerItem, QToolButton, QTreeWidgetItem,
    QVBoxLayout, QWidget)

from ..widgets.parameter_tree_widget import ParameterTreeWidget
from spine_items.data_transformer.widgets.class_tree_widget import ClassTreeWidget
from spine_items.data_transformer.widgets.drop_target_table import DropTargetTable
from spinetoolbox.widgets.custom_combobox import ElidedCombobox
from spine_items import resources_icons_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(800, 735)
        self.remove_parameter_rename_action = QAction(MainWindow)
        self.remove_parameter_rename_action.setObjectName(u"remove_parameter_rename_action")
        self.remove_parameter_rename_action.setShortcutContext(Qt.WidgetShortcut)
        self.remove_value_transformation_action = QAction(MainWindow)
        self.remove_value_transformation_action.setObjectName(u"remove_value_transformation_action")
        self.remove_value_transformation_action.setShortcutContext(Qt.WidgetShortcut)
        self.remove_class_rename_action = QAction(MainWindow)
        self.remove_class_rename_action.setObjectName(u"remove_class_rename_action")
        self.remove_class_rename_action.setShortcutContext(Qt.WidgetShortcut)
        self.remove_instruction_action = QAction(MainWindow)
        self.remove_instruction_action.setObjectName(u"remove_instruction_action")
        self.remove_instruction_action.setShortcutContext(Qt.WidgetShortcut)
        self.copy_parameter_rename_data_action = QAction(MainWindow)
        self.copy_parameter_rename_data_action.setObjectName(u"copy_parameter_rename_data_action")
        self.copy_parameter_rename_data_action.setShortcutContext(Qt.WidgetShortcut)
        self.paste_parameter_rename_data_action = QAction(MainWindow)
        self.paste_parameter_rename_data_action.setObjectName(u"paste_parameter_rename_data_action")
        self.paste_parameter_rename_data_action.setShortcutContext(Qt.WidgetShortcut)
        self.copy_class_rename_data_action = QAction(MainWindow)
        self.copy_class_rename_data_action.setObjectName(u"copy_class_rename_data_action")
        self.copy_class_rename_data_action.setShortcutContext(Qt.WidgetShortcut)
        self.paste_class_rename_data_action = QAction(MainWindow)
        self.paste_class_rename_data_action.setObjectName(u"paste_class_rename_data_action")
        self.paste_class_rename_data_action.setShortcutContext(Qt.WidgetShortcut)
        self.copy_value_transformation_data_action = QAction(MainWindow)
        self.copy_value_transformation_data_action.setObjectName(u"copy_value_transformation_data_action")
        self.copy_value_transformation_data_action.setShortcutContext(Qt.WidgetShortcut)
        self.paste_value_transformation_data_action = QAction(MainWindow)
        self.paste_value_transformation_data_action.setObjectName(u"paste_value_transformation_data_action")
        self.paste_value_transformation_data_action.setShortcutContext(Qt.WidgetShortcut)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        MainWindow.setCentralWidget(self.centralwidget)
        self.type_dock = QDockWidget(MainWindow)
        self.type_dock.setObjectName(u"type_dock")
        self.type_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents = QWidget()
        self.dockWidgetContents.setObjectName(u"dockWidgetContents")
        self.verticalLayout = QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout.setSpacing(3)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.label = QLabel(self.dockWidgetContents)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setBold(True)
        self.label.setFont(font)

        self.horizontalLayout_3.addWidget(self.label)

        self.filter_combo_box = QComboBox(self.dockWidgetContents)
        self.filter_combo_box.setObjectName(u"filter_combo_box")

        self.horizontalLayout_3.addWidget(self.filter_combo_box)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_2)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.type_dock.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(Qt.LeftDockWidgetArea, self.type_dock)
        self.load_database_dock = QDockWidget(MainWindow)
        self.load_database_dock.setObjectName(u"load_database_dock")
        self.load_database_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents_2 = QWidget()
        self.dockWidgetContents_2.setObjectName(u"dockWidgetContents_2")
        self.verticalLayout_2 = QVBoxLayout(self.dockWidgetContents_2)
        self.verticalLayout_2.setSpacing(3)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_4 = QLabel(self.dockWidgetContents_2)
        self.label_4.setObjectName(u"label_4")

        self.horizontalLayout_4.addWidget(self.label_4)

        self.database_url_combo_box = ElidedCombobox(self.dockWidgetContents_2)
        self.database_url_combo_box.setObjectName(u"database_url_combo_box")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.database_url_combo_box.sizePolicy().hasHeightForWidth())
        self.database_url_combo_box.setSizePolicy(sizePolicy)
        self.database_url_combo_box.setSizeAdjustPolicy(QComboBox.AdjustToMinimumContentsLengthWithIcon)
        self.database_url_combo_box.setMinimumContentsLength(0)

        self.horizontalLayout_4.addWidget(self.database_url_combo_box)

        self.load_url_from_fs_button = QToolButton(self.dockWidgetContents_2)
        self.load_url_from_fs_button.setObjectName(u"load_url_from_fs_button")
        icon = QIcon()
        icon.addFile(u":/icons/file.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.load_url_from_fs_button.setIcon(icon)

        self.horizontalLayout_4.addWidget(self.load_url_from_fs_button)


        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_3)

        self.load_data_button = QPushButton(self.dockWidgetContents_2)
        self.load_data_button.setObjectName(u"load_data_button")
        self.load_data_button.setEnabled(False)

        self.horizontalLayout_5.addWidget(self.load_data_button)


        self.verticalLayout_2.addLayout(self.horizontalLayout_5)

        self.verticalSpacer_2 = QSpacerItem(20, 2, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer_2)

        self.load_database_dock.setWidget(self.dockWidgetContents_2)
        MainWindow.addDockWidget(Qt.LeftDockWidgetArea, self.load_database_dock)
        self.possible_parameters_dock = QDockWidget(MainWindow)
        self.possible_parameters_dock.setObjectName(u"possible_parameters_dock")
        self.possible_parameters_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents_3 = QWidget()
        self.dockWidgetContents_3.setObjectName(u"dockWidgetContents_3")
        self.verticalLayout_3 = QVBoxLayout(self.dockWidgetContents_3)
        self.verticalLayout_3.setSpacing(3)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(3, 3, 3, 3)
        self.available_parameters_tree_view = ParameterTreeWidget(self.dockWidgetContents_3)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, u"1");
        self.available_parameters_tree_view.setHeaderItem(__qtreewidgetitem)
        self.available_parameters_tree_view.setObjectName(u"available_parameters_tree_view")
        self.available_parameters_tree_view.setDragEnabled(True)
        self.available_parameters_tree_view.setDragDropMode(QAbstractItemView.DragOnly)
        self.available_parameters_tree_view.setDefaultDropAction(Qt.CopyAction)
        self.available_parameters_tree_view.header().setVisible(False)

        self.verticalLayout_3.addWidget(self.available_parameters_tree_view)

        self.possible_parameters_dock.setWidget(self.dockWidgetContents_3)
        MainWindow.addDockWidget(Qt.LeftDockWidgetArea, self.possible_parameters_dock)
        self.parameter_rename_dock = QDockWidget(MainWindow)
        self.parameter_rename_dock.setObjectName(u"parameter_rename_dock")
        self.parameter_rename_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents_4 = QWidget()
        self.dockWidgetContents_4.setObjectName(u"dockWidgetContents_4")
        self.verticalLayout_4 = QVBoxLayout(self.dockWidgetContents_4)
        self.verticalLayout_4.setSpacing(3)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.add_parameter_button = QPushButton(self.dockWidgetContents_4)
        self.add_parameter_button.setObjectName(u"add_parameter_button")

        self.horizontalLayout.addWidget(self.add_parameter_button)

        self.remove_parameter_button = QPushButton(self.dockWidgetContents_4)
        self.remove_parameter_button.setObjectName(u"remove_parameter_button")

        self.horizontalLayout.addWidget(self.remove_parameter_button)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.verticalLayout_4.addLayout(self.horizontalLayout)

        self.parameter_rename_table_view = DropTargetTable(self.dockWidgetContents_4)
        self.parameter_rename_table_view.setObjectName(u"parameter_rename_table_view")
        self.parameter_rename_table_view.setDragDropMode(QAbstractItemView.DropOnly)
        self.parameter_rename_table_view.setSortingEnabled(True)
        self.parameter_rename_table_view.horizontalHeader().setStretchLastSection(True)

        self.verticalLayout_4.addWidget(self.parameter_rename_table_view)

        self.parameter_rename_dock.setWidget(self.dockWidgetContents_4)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.parameter_rename_dock)
        self.value_transformation_dock = QDockWidget(MainWindow)
        self.value_transformation_dock.setObjectName(u"value_transformation_dock")
        self.value_transformation_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents_5 = QWidget()
        self.dockWidgetContents_5.setObjectName(u"dockWidgetContents_5")
        self.verticalLayout_5 = QVBoxLayout(self.dockWidgetContents_5)
        self.verticalLayout_5.setSpacing(3)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.add_transformation_button = QPushButton(self.dockWidgetContents_5)
        self.add_transformation_button.setObjectName(u"add_transformation_button")

        self.horizontalLayout_2.addWidget(self.add_transformation_button)

        self.remove_transformation_button = QPushButton(self.dockWidgetContents_5)
        self.remove_transformation_button.setObjectName(u"remove_transformation_button")

        self.horizontalLayout_2.addWidget(self.remove_transformation_button)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_4)


        self.verticalLayout_5.addLayout(self.horizontalLayout_2)

        self.transformations_table_view = DropTargetTable(self.dockWidgetContents_5)
        self.transformations_table_view.setObjectName(u"transformations_table_view")
        self.transformations_table_view.setAcceptDrops(True)
        self.transformations_table_view.setDragDropMode(QAbstractItemView.DropOnly)
        self.transformations_table_view.setShowGrid(False)
        self.transformations_table_view.setSortingEnabled(True)
        self.transformations_table_view.horizontalHeader().setStretchLastSection(True)

        self.verticalLayout_5.addWidget(self.transformations_table_view)

        self.value_transformation_dock.setWidget(self.dockWidgetContents_5)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.value_transformation_dock)
        self.value_instructions_dock = QDockWidget(MainWindow)
        self.value_instructions_dock.setObjectName(u"value_instructions_dock")
        self.value_instructions_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents_6 = QWidget()
        self.dockWidgetContents_6.setObjectName(u"dockWidgetContents_6")
        self.verticalLayout_6 = QVBoxLayout(self.dockWidgetContents_6)
        self.verticalLayout_6.setSpacing(3)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.add_instruction_button = QPushButton(self.dockWidgetContents_6)
        self.add_instruction_button.setObjectName(u"add_instruction_button")
        self.add_instruction_button.setEnabled(False)

        self.horizontalLayout_6.addWidget(self.add_instruction_button)

        self.remove_instruction_button = QPushButton(self.dockWidgetContents_6)
        self.remove_instruction_button.setObjectName(u"remove_instruction_button")
        self.remove_instruction_button.setEnabled(False)

        self.horizontalLayout_6.addWidget(self.remove_instruction_button)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_5)


        self.verticalLayout_6.addLayout(self.horizontalLayout_6)

        self.instructions_list_view = QListWidget(self.dockWidgetContents_6)
        self.instructions_list_view.setObjectName(u"instructions_list_view")
        self.instructions_list_view.setEnabled(False)

        self.verticalLayout_6.addWidget(self.instructions_list_view)

        self.instruction_options_layout = QFormLayout()
        self.instruction_options_layout.setObjectName(u"instruction_options_layout")
        self.label_2 = QLabel(self.dockWidgetContents_6)
        self.label_2.setObjectName(u"label_2")

        self.instruction_options_layout.setWidget(0, QFormLayout.LabelRole, self.label_2)

        self.operation_combo_box = QComboBox(self.dockWidgetContents_6)
        self.operation_combo_box.setObjectName(u"operation_combo_box")

        self.instruction_options_layout.setWidget(0, QFormLayout.FieldRole, self.operation_combo_box)


        self.verticalLayout_6.addLayout(self.instruction_options_layout)

        self.value_instructions_dock.setWidget(self.dockWidgetContents_6)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.value_instructions_dock)
        self.possible_classes_dock = QDockWidget(MainWindow)
        self.possible_classes_dock.setObjectName(u"possible_classes_dock")
        self.possible_classes_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents_7 = QWidget()
        self.dockWidgetContents_7.setObjectName(u"dockWidgetContents_7")
        self.verticalLayout_7 = QVBoxLayout(self.dockWidgetContents_7)
        self.verticalLayout_7.setSpacing(3)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(3, 3, 3, 3)
        self.available_classes_tree_widget = ClassTreeWidget(self.dockWidgetContents_7)
        __qtreewidgetitem1 = QTreeWidgetItem()
        __qtreewidgetitem1.setText(0, u"1");
        self.available_classes_tree_widget.setHeaderItem(__qtreewidgetitem1)
        self.available_classes_tree_widget.setObjectName(u"available_classes_tree_widget")
        self.available_classes_tree_widget.setDragEnabled(True)
        self.available_classes_tree_widget.setDragDropMode(QAbstractItemView.DragOnly)
        self.available_classes_tree_widget.setDefaultDropAction(Qt.CopyAction)
        self.available_classes_tree_widget.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.available_classes_tree_widget.header().setVisible(False)

        self.verticalLayout_7.addWidget(self.available_classes_tree_widget)

        self.possible_classes_dock.setWidget(self.dockWidgetContents_7)
        MainWindow.addDockWidget(Qt.LeftDockWidgetArea, self.possible_classes_dock)
        self.class_rename_dock = QDockWidget(MainWindow)
        self.class_rename_dock.setObjectName(u"class_rename_dock")
        self.class_rename_dock.setFeatures(QDockWidget.DockWidgetMovable)
        self.dockWidgetContents_8 = QWidget()
        self.dockWidgetContents_8.setObjectName(u"dockWidgetContents_8")
        self.verticalLayout_8 = QVBoxLayout(self.dockWidgetContents_8)
        self.verticalLayout_8.setSpacing(3)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.verticalLayout_8.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout_7 = QHBoxLayout()
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.add_class_button = QPushButton(self.dockWidgetContents_8)
        self.add_class_button.setObjectName(u"add_class_button")

        self.horizontalLayout_7.addWidget(self.add_class_button)

        self.remove_class_button = QPushButton(self.dockWidgetContents_8)
        self.remove_class_button.setObjectName(u"remove_class_button")

        self.horizontalLayout_7.addWidget(self.remove_class_button)

        self.horizontalSpacer_6 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_7.addItem(self.horizontalSpacer_6)


        self.verticalLayout_8.addLayout(self.horizontalLayout_7)

        self.class_rename_table_view = DropTargetTable(self.dockWidgetContents_8)
        self.class_rename_table_view.setObjectName(u"class_rename_table_view")
        self.class_rename_table_view.setAcceptDrops(True)
        self.class_rename_table_view.setDragDropMode(QAbstractItemView.DropOnly)
        self.class_rename_table_view.setSortingEnabled(True)
        self.class_rename_table_view.horizontalHeader().setStretchLastSection(True)

        self.verticalLayout_8.addWidget(self.class_rename_table_view)

        self.class_rename_dock.setWidget(self.dockWidgetContents_8)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.class_rename_dock)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.remove_parameter_rename_action.setText(QCoreApplication.translate("MainWindow", u"Remove parameter", None))
#if QT_CONFIG(shortcut)
        self.remove_parameter_rename_action.setShortcut(QCoreApplication.translate("MainWindow", u"Del", None))
#endif // QT_CONFIG(shortcut)
        self.remove_value_transformation_action.setText(QCoreApplication.translate("MainWindow", u"Remove transformation", None))
#if QT_CONFIG(shortcut)
        self.remove_value_transformation_action.setShortcut(QCoreApplication.translate("MainWindow", u"Del", None))
#endif // QT_CONFIG(shortcut)
        self.remove_class_rename_action.setText(QCoreApplication.translate("MainWindow", u"Remove class", None))
#if QT_CONFIG(shortcut)
        self.remove_class_rename_action.setShortcut(QCoreApplication.translate("MainWindow", u"Del", None))
#endif // QT_CONFIG(shortcut)
        self.remove_instruction_action.setText(QCoreApplication.translate("MainWindow", u"Remove instruction", None))
#if QT_CONFIG(shortcut)
        self.remove_instruction_action.setShortcut(QCoreApplication.translate("MainWindow", u"Del", None))
#endif // QT_CONFIG(shortcut)
        self.copy_parameter_rename_data_action.setText(QCoreApplication.translate("MainWindow", u"Copy", None))
#if QT_CONFIG(shortcut)
        self.copy_parameter_rename_data_action.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+C", None))
#endif // QT_CONFIG(shortcut)
        self.paste_parameter_rename_data_action.setText(QCoreApplication.translate("MainWindow", u"Paste", None))
#if QT_CONFIG(shortcut)
        self.paste_parameter_rename_data_action.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+V", None))
#endif // QT_CONFIG(shortcut)
        self.copy_class_rename_data_action.setText(QCoreApplication.translate("MainWindow", u"Copy", None))
#if QT_CONFIG(shortcut)
        self.copy_class_rename_data_action.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+C", None))
#endif // QT_CONFIG(shortcut)
        self.paste_class_rename_data_action.setText(QCoreApplication.translate("MainWindow", u"Paste", None))
#if QT_CONFIG(shortcut)
        self.paste_class_rename_data_action.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+V", None))
#endif // QT_CONFIG(shortcut)
        self.copy_value_transformation_data_action.setText(QCoreApplication.translate("MainWindow", u"Copy", None))
#if QT_CONFIG(shortcut)
        self.copy_value_transformation_data_action.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+C", None))
#endif // QT_CONFIG(shortcut)
        self.paste_value_transformation_data_action.setText(QCoreApplication.translate("MainWindow", u"Paste", None))
#if QT_CONFIG(shortcut)
        self.paste_value_transformation_data_action.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+V", None))
#endif // QT_CONFIG(shortcut)
        self.type_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Select transformation", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Transformation type:", None))
        self.load_database_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Load from database", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Database url:", None))
#if QT_CONFIG(tooltip)
        self.load_url_from_fs_button.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Browse file system</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.load_url_from_fs_button.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.load_data_button.setText(QCoreApplication.translate("MainWindow", u"Load filter data from database", None))
        self.possible_parameters_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Possible parameters", None))
        self.parameter_rename_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Rename parameters", None))
        self.add_parameter_button.setText(QCoreApplication.translate("MainWindow", u"Add", None))
        self.remove_parameter_button.setText(QCoreApplication.translate("MainWindow", u"Remove", None))
        self.value_transformation_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Parameter value transformation", None))
        self.add_transformation_button.setText(QCoreApplication.translate("MainWindow", u"Add", None))
        self.remove_transformation_button.setText(QCoreApplication.translate("MainWindow", u"Remove", None))
        self.value_instructions_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Transformation instructions", None))
        self.add_instruction_button.setText(QCoreApplication.translate("MainWindow", u"Add", None))
        self.remove_instruction_button.setText(QCoreApplication.translate("MainWindow", u"Remove", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Operation:", None))
        self.possible_classes_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Possible classes", None))
        self.class_rename_dock.setWindowTitle(QCoreApplication.translate("MainWindow", u"Rename classes", None))
        self.add_class_button.setText(QCoreApplication.translate("MainWindow", u"Add", None))
        self.remove_class_button.setText(QCoreApplication.translate("MainWindow", u"Remove", None))
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Module for data connection class."""
import logging
import os
import pathlib
import shutil
from PySide6.QtCore import QFileInfo, QItemSelection, QModelIndex, Qt, QTimer, Slot
from PySide6.QtGui import QBrush, QStandardItem, QStandardItemModel
from PySide6.QtWidgets import QFileDialog, QFileIconProvider, QGraphicsItem, QInputDialog, QMessageBox
from spine_engine.utils.serialization import deserialize_path, serialize_path
from spinedb_api.helpers import remove_credentials_from_url
from spinetoolbox.config import INVALID_FILENAME_CHARS
from spinetoolbox.helpers import open_url, same_path
from spinetoolbox.project_item.project_item import ProjectItem
from spinetoolbox.widgets.custom_qwidgets import ToolBarWidget
from ..database_validation import DatabaseConnectionValidator
from ..utils import convert_to_sqlalchemy_url, convert_url_to_safe_string
from ..widgets import UrlSelectorDialog
from .commands import AddDCReferencesCommand, MoveReferenceToData, RemoveDCReferencesCommand
from .custom_file_system_watcher import CustomFileSystemWatcher
from .executable_item import ExecutableItem
from .item_info import ItemInfo
from .output_resources import scan_for_resources
from .utils import restore_database_references


class _Role:
    """Extra reference model data roles."""

    DATA_FILE_PATH = Qt.ItemDataRole.UserRole + 1
    MISSING = Qt.ItemDataRole.UserRole + 2
    FILE_REFERENCE = Qt.ItemDataRole.UserRole + 3
    DB_URL_REFERENCE = Qt.ItemDataRole.UserRole + 4


_MISSING_ITEM_FOREGROUND = QBrush(Qt.GlobalColor.red)


class DataConnection(ProjectItem):
    """Data connection project item."""

    def __init__(self, name, description, x, y, toolbox, project, file_references=None, db_references=None):
        """
        Args:
            name (str): Object name
            description (str): Object description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (ToolboxUI): QMainWindow instance
            project (SpineToolboxProject): the project this item belongs to
            file_references (list, optional): a list of file paths
            db_references (list of dict, optional): a list of db urls
        """
        super().__init__(name, description, x, y, project)
        if file_references is None:
            file_references = []
        if db_references is None:
            db_references = []
        self._toolbox = toolbox
        self.reference_model = QStandardItemModel()  # References
        self.data_model = QStandardItemModel()  # Paths of project internal files. These are found in DC data directory
        self.file_system_watcher = None
        self.file_references = list(file_references)
        self._file_ref_root = QStandardItem("File paths")
        self._file_ref_root.setFlags(self._file_ref_root.flags() & ~Qt.ItemIsEditable)
        self._db_ref_root = QStandardItem("URLs")
        self._db_ref_root.setFlags(self._file_ref_root.flags() & ~Qt.ItemIsEditable)
        self.file_refs_selected = False
        self.any_refs_selected = False
        self.any_data_selected = False
        self.current_is_file_ref = False
        self.populate_reference_list(db_references)
        self.populate_data_list()
        self._database_validator = DatabaseConnectionValidator()

    def set_up(self):
        super().set_up()
        self.file_system_watcher = CustomFileSystemWatcher(self)
        self.file_system_watcher.add_persistent_file_paths(ref for ref in self.file_references if os.path.exists(ref))
        self._watch_sqlite_file(*self.db_reference_iter())
        self.file_system_watcher.add_persistent_dir_path(self.data_dir)
        self.file_system_watcher.file_removed.connect(self._handle_file_removed)
        self.file_system_watcher.file_renamed.connect(self._handle_file_renamed)
        self.file_system_watcher.file_added.connect(self._handle_file_added)

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def db_reference_iter(self):
        """Iterates over database references.

        Yields:
            dict: database URL
        """
        for row in range(self._db_ref_root.rowCount()):
            yield self._db_ref_root.child(row).data(_Role.DB_URL_REFERENCE)

    def has_db_references(self):
        return self._db_ref_root.rowCount() != 0

    @Slot(QItemSelection, QItemSelection)
    def _update_selection_state(self, _selected, _deselected):
        self._do_update_selection_state()

    def _do_update_selection_state(self):
        ref_indexes = self._properties_ui.treeView_dc_references.selectionModel().selectedIndexes()
        data_indexes = self._properties_ui.treeView_dc_data.selectionModel().selectedIndexes()
        self.file_refs_selected = any(ind.parent().row() == 0 for ind in ref_indexes)
        self.any_refs_selected = any(ind.parent().row() in (0, 1) for ind in ref_indexes)
        self.current_is_file_ref = self._properties_ui.treeView_dc_references.currentIndex().parent().row() == 0
        self.any_data_selected = bool(data_indexes)
        self._properties_ui.toolButton_minus.setEnabled(self.any_refs_selected)
        self._properties_ui.toolButton_add.setEnabled(self.file_refs_selected)

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.toolButton_minus.clicked] = self.remove_references
        s[self._properties_ui.toolButton_add.clicked] = self.copy_to_project
        s[self._properties_ui.treeView_dc_references.doubleClicked] = self.open_reference
        s[self._properties_ui.treeView_dc_data.doubleClicked] = self.open_data_file
        s[self._properties_ui.treeView_dc_references.files_dropped] = self._add_file_references
        s[self._properties_ui.treeView_dc_data.files_dropped] = self.add_data_files
        s[self.get_icon().files_dropped_on_icon] = self.receive_files_dropped_on_icon
        s[self._properties_ui.treeView_dc_references.del_key_pressed] = lambda b=False: self.remove_references(b)
        s[self._properties_ui.treeView_dc_data.del_key_pressed] = self.remove_files
        s[self._properties_ui.treeView_dc_references.selectionModel().selectionChanged] = self._update_selection_state
        s[self._properties_ui.treeView_dc_data.selectionModel().selectionChanged] = self._update_selection_state
        s[self._properties_ui.action_new_file_reference.triggered] = self.show_add_file_references_dialog
        s[self._properties_ui.action_new_db_reference.triggered] = self.show_add_db_reference_dialog
        return s

    def restore_selections(self):
        """Restore selections into shared widgets when this project item is selected."""
        self._properties_ui.treeView_dc_references.setModel(self.reference_model)
        self._properties_ui.treeView_dc_data.setModel(self.data_model)
        self._do_update_selection_state()
        self._properties_ui.treeView_dc_references.expandAll()
        file_ref_root_widget = ToolBarWidget("File paths")
        file_ref_root_widget.tool_bar.setToolButtonStyle(Qt.ToolButtonIconOnly)
        file_ref_root_widget.tool_bar.addAction(self._properties_ui.action_new_file_reference)
        self._properties_ui.treeView_dc_references.setIndexWidget(
            self.reference_model.indexFromItem(self._file_ref_root), file_ref_root_widget
        )
        db_ref_root_widget = ToolBarWidget("URLs")
        db_ref_root_widget.tool_bar.setToolButtonStyle(Qt.ToolButtonIconOnly)
        db_ref_root_widget.tool_bar.addAction(self._properties_ui.action_new_db_reference)
        self._properties_ui.treeView_dc_references.setIndexWidget(
            self.reference_model.indexFromItem(self._db_ref_root), db_ref_root_widget
        )

    @Slot(QGraphicsItem, list)
    def receive_files_dropped_on_icon(self, icon, file_paths):
        """Called when files are dropped onto a data connection graphics item.
        If the item is this Data Connection's graphics item, add the files to data."""
        if icon == self.get_icon():
            self.add_data_files(file_paths)

    @Slot(list)
    def add_data_files(self, file_paths):
        """Add files to data directory"""
        for file_path in file_paths:
            filename = os.path.split(file_path)[1]
            self._logger.msg.emit(f"Copying file <b>{filename}</b> to <b>{self.name}</b>")
            try:
                shutil.copy(file_path, self.data_dir)
            except OSError:
                self._logger.msg_error.emit("[OSError] Copying failed")

    @Slot(bool)
    def show_add_file_references_dialog(self, _=False):
        """Opens a file browser where user can select files to be added as references for this Data Connection."""
        answer = QFileDialog.getOpenFileNames(self._toolbox, "Add file references", self._project.project_dir, "*.*")
        file_paths = answer[0]
        if not file_paths:  # Cancel button clicked
            return
        self._add_file_references(file_paths)

    @Slot(list)
    def _add_file_references(self, paths):
        """Add multiple file paths to reference list.

        Args:
            paths (list): A list of paths to files
        """
        repeated_paths = []
        new_paths = []
        existing_references = (reference for reference in self.file_references if os.path.isfile(reference))
        for path in paths:
            if not os.path.isfile(path):
                continue
            if any(os.path.samefile(path, ref) for ref in existing_references):
                repeated_paths.append(path)
            else:
                new_paths.append(path)
        repeated_paths = ", ".join(repeated_paths)
        if repeated_paths:
            self._logger.msg_warning.emit(f"Reference to file(s) <b>{repeated_paths}</b> already exists")
        if new_paths:
            self._toolbox.undo_stack.push(AddDCReferencesCommand(self.name, new_paths, [], self._project))

    @Slot(bool)
    def show_add_db_reference_dialog(self, _=False):
        """Opens a dialog where user can select a url to be added as reference for this Data Connection."""
        selector = UrlSelectorDialog(self._toolbox.qsettings(), False, self._toolbox, self._toolbox)
        result = selector.exec()
        if result == UrlSelectorDialog.DialogCode.Rejected:
            return
        url = selector.url_dict()
        if self._has_db_reference(url):
            self._logger.msg_warning.emit(f"Reference to database <b>{url}</b> already exists")
            return
        sa_url = convert_to_sqlalchemy_url(url, self.name, self._logger)
        self._database_validator.validate_url(
            url["dialect"], sa_url, self._log_database_reference_error, success_slot=None
        )
        self._toolbox.undo_stack.push(AddDCReferencesCommand(self.name, [], [url], self._project))

    def _has_db_reference(self, url):
        """Checks if given database URL exists already.

        Ignores usernames and passwords.

        Args:
            url (dict): URL to check

        Returns:
            bool: True if db reference exists, False otherwise
        """
        significant_keys = ("dialect", "host", "port", "database")
        for row in range(self._db_ref_root.rowCount()):
            existing_url = self._db_ref_root.child(row).data(_Role.DB_URL_REFERENCE)
            if all(url[key] == existing_url[key] for key in significant_keys):
                return True
        return False

    @Slot(str, object)
    def _log_database_reference_error(self, error, url):
        """Logs final database validation error messages.

        Args:
            error (str): message
            url (URL): SqlAlchemy URL of the database
        """
        url_text = remove_credentials_from_url(str(url))
        for row in range(self._db_ref_root.rowCount()):
            item = self._db_ref_root.child(row)
            if url_text == item.text():
                self._mark_as_missing(item)
                break
        self._logger.msg_error.emit(f"<b>{self.name}</b>: invalid database URL: {error}")

    def do_add_references(self, file_refs, db_refs):
        """Adds file and databases references to DC and starts watching the files.

        Args:
            file_refs (list of str): file reference paths
            db_refs (list of dict): database reference URLs
        """
        file_refs = [os.path.abspath(ref) for ref in file_refs]
        self.file_references += file_refs
        self.file_system_watcher.add_persistent_file_paths(ref for ref in file_refs if os.path.exists(ref))
        self._append_file_references_to_model(*file_refs)
        self._watch_sqlite_file(*db_refs)
        self._append_db_references_to_model(*db_refs)
        self._check_notifications()
        self._resources_to_successors_changed()

    @Slot(bool)
    def remove_references(self, _=False):
        """Pushes a remove references command to undo stack"""
        indexes = self._properties_ui.treeView_dc_references.selectedIndexes()
        if not indexes:  # Nothing selected
            self._logger.msg.emit("Please select references to remove")
            return
        file_ref_root_index = self.reference_model.indexFromItem(self._file_ref_root)
        db_ref_root_index = self.reference_model.indexFromItem(self._db_ref_root)
        file_references = []
        db_references = []
        for index in indexes:
            parent = index.parent()
            if not parent.isValid():
                continue
            if parent == file_ref_root_index:
                file_references.append(index.data(Qt.ItemDataRole.DisplayRole))
            elif parent == db_ref_root_index:
                db_references.append(index.data(_Role.DB_URL_REFERENCE))
        self._toolbox.undo_stack.push(
            RemoveDCReferencesCommand(self.name, file_references, db_references, self._project)
        )
        self._logger.msg.emit("Selected references removed")

    def do_remove_references(self, file_refs, db_refs):
        """Removes given paths from references.

        Args:
            file_refs (list of str): List of removed file paths.
            db_refs (list of dict): List of removed urls.
        """
        self.file_system_watcher.remove_persistent_file_paths(file_refs)
        self._unwatch_sqlite_file(*db_refs)
        refs_removed = self._remove_file_references(*file_refs)
        refs_removed |= self._remove_db_references(*db_refs)
        if refs_removed:
            self._check_notifications()
            self._resources_to_successors_changed()

    def _remove_file_references(self, *refs):
        result = False
        for k in reversed(range(self._file_ref_root.rowCount())):
            if any(same_path(self._file_ref_root.child(k).text(), ref) for ref in refs):
                self.file_references.pop(k)
                self._file_ref_root.removeRow(k)
                result = True
        return result

    def _remove_db_references(self, *refs):
        result = False
        matches = {convert_url_to_safe_string(url) for url in refs}
        for k in reversed(range(self._db_ref_root.rowCount())):
            if self._db_ref_root.child(k).text() in matches:
                self._db_ref_root.removeRow(k)
                result = True
        return result

    def _remove_data_file(self, path):
        for k in reversed(range(self.data_model.rowCount())):
            data_filepath = self.data_model.item(k).data(_Role.DATA_FILE_PATH)
            if same_path(data_filepath, path):
                self.data_model.removeRow(k)
                return True
        return False

    def _rename_data_file(self, old_path, new_path):
        for k in range(self.data_model.rowCount()):
            item = self.data_model.item(k)
            if same_path(item.data(_Role.DATA_FILE_PATH), old_path):
                item.setText(os.path.basename(new_path))
                item.setData(new_path, _Role.DATA_FILE_PATH)
                return True
        return False

    def copy_local_data(self, item_dict):
        """See base class."""
        original_data_dir = item_dict.get("original_data_dir")  # (str) original dir of duplicated ProjectItem
        duplicate_files = item_dict.get("duplicate_files")  # (bool) Flag indicating if linked files should be copied
        if not original_data_dir and not duplicate_files:
            return
        if duplicate_files:
            try:
                os.makedirs(self.data_dir, exist_ok=True)  # Make data dir
            except OSError as e:
                self._logger.msg_error.emit(f"Creating directory {self.data_dir} failed. {e}.")
                return
            for f in os.listdir(original_data_dir):
                src_file = os.path.join(original_data_dir, f)
                if os.path.isfile(src_file):
                    dst_file = os.path.join(self.data_dir, f)
                    try:
                        shutil.copy(src_file, dst_file)  # Copy file
                    except OSError as e:
                        self._logger.msg_error.emit(f"Copying file <b>{f}</b> to <b>{self.data_dir}</b> failed. [{e}]")
                        continue
                    self._logger.msg.emit(f"File <b>{f}</b> copied to <b>{self.data_dir}</b>")
        self.populate_data_list()

    @Slot(str)
    def _handle_file_removed(self, path):
        """Marks file reference missing or removes data reference.

        Args:
            path (str): file path
        """
        resources_changed = self._try_to_mark_file_reference_missing(path) or self._remove_data_file(path)
        if resources_changed:
            self._check_notifications()
            self._resources_to_successors_changed()

    def _try_to_mark_file_reference_missing(self, path):
        """Marks a file reference as missing if given path is in references.

        Args:
            path (str): file path

        Returns:
            bool: True if references was marked missing successfully, False otherwise
        """
        for row in range(self._file_ref_root.rowCount()):
            item = self._file_ref_root.child(row)
            if same_path(item.text(), path):
                self._mark_as_missing(item)
                return True
        for row in range(self._db_ref_root.rowCount()):
            item = self._db_ref_root.child(row)
            url = item.data(_Role.DB_URL_REFERENCE)
            if url["dialect"] != "sqlite":
                continue
            if same_path(url["database"], path):
                self._mark_as_missing(item)
                return True
        return False

    @Slot(str, str)
    def _handle_file_renamed(self, old_path, new_path):
        """Marks file reference missing or renames data file.

        Args:
            old_path (str): original path
            new_path (str): renamed path
        """

        def replace_new_path(paths):
            for i, path in enumerate(paths):
                if path == new_path:
                    paths[i] = old_path
                    return True
            return False

        if self._try_to_mark_file_reference_missing(old_path):
            self._refresh_file_reference_delayed(old_path)
            self._check_notifications()
            self._resources_to_successors_changed()
            return
        if not self._rename_data_file(old_path, new_path):
            return
        self._check_notifications()
        file_refs = list(self.file_references)
        data_files = [os.path.join(self.data_dir, f) for f in self.data_files()]
        new_resources = scan_for_resources(
            self, file_refs + data_files, list(self.db_reference_iter()), self._project.project_dir
        )
        if not replace_new_path(file_refs):
            replace_new_path(data_files)
        old_resources = scan_for_resources(
            self, file_refs + data_files, list(self.db_reference_iter()), self._project.project_dir
        )
        self._resources_to_successors_replaced(old_resources, new_resources)

    def _refresh_file_reference_delayed(self, path):
        """Checks if given file exists after certain delay.

        Args:
            path (str): file path
        """

        # Some software saves files by renaming them and then creating
        # a new file with the same name for safety reasons.
        # So, sometimes a file goes "missing" for a moment and then comes back.
        # We'll try to refresh the file reference a couple times in this case.
        def refresh(retry_count):
            if not os.path.exists(path):
                if retry_count < 5:
                    QTimer.singleShot(1000, lambda: refresh(retry_count + 1))
                return
            fixed_references = []
            for row in range(self._file_ref_root.rowCount()):
                item = self._file_ref_root.child(row, 0)
                if not same_path(path, item.text()):
                    continue
                fixed_references.append(path)
                self._mark_as_found(item)
            if not fixed_references:
                return
            self.file_system_watcher.add_persistent_file_paths(ref for ref in fixed_references)
            self._check_notifications()
            self._resources_to_successors_changed()

        QTimer.singleShot(1000, lambda: refresh(0))

    @Slot(str)
    def _handle_file_added(self, path):
        if same_path(os.path.dirname(path), self.data_dir):
            self._append_data_files_to_model(path)
            self._check_notifications()
            self._resources_to_successors_changed()

    @Slot(bool)
    def copy_to_project(self, _=False):
        """Copy selected file references to this Data Connection's data directory."""
        selected_indexes = self._properties_ui.treeView_dc_references.selectedIndexes()
        if not selected_indexes:
            self._logger.msg_warning.emit("No files to copy")
            return
        self._toolbox.undo_stack.push(
            MoveReferenceToData(self.name, [index.data() for index in selected_indexes], self._project)
        )

    def do_copy_to_project(self, paths):
        """Copies given files to item's data directory.

        Args:
            paths (Iterable of str): paths to copy
        """
        for path in paths:
            if not os.path.exists(path):
                self._logger.msg_error.emit(f"File <b>{path}</b> does not exist")
                continue
            filename = os.path.basename(path)
            self._logger.msg.emit(f"Copying file <b>{filename}</b> to Data Connection <b>{self.name}</b>")
            try:
                shutil.copy(path, self.data_dir)
            except OSError:
                self._logger.msg_error.emit("[OSError] Copying failed")

    def refresh_references(self):
        """Checks if missing file references have somehow come back to life."""
        selected_indexes = self._properties_ui.treeView_dc_references.selectedIndexes()
        if not selected_indexes:
            return
        for index in selected_indexes:
            item = self.reference_model.itemFromIndex(index)
            if self.reference_model.itemFromIndex(index.parent()) is self._db_ref_root:
                self.refresh_db_references(item)
            else:
                self.refresh_file_references(item)

    def refresh_file_references(self, item):
        file_path = item.data(Qt.ItemDataRole.DisplayRole)
        if item.data(_Role.MISSING) and os.path.exists(file_path):
            self._mark_as_found(item)
        else:
            return
        self.file_system_watcher.add_persistent_file_path(file_path)
        self._check_notifications()
        self._resources_to_successors_changed()

    def refresh_db_references(self, item):
        """Checks if the db reference is valid"""
        url = item.data(_Role.DB_URL_REFERENCE)
        self._database_validator.validate_url(
            url["dialect"],
            convert_to_sqlalchemy_url(url),
            self._log_database_reference_error,
            success_slot=self._revive_db_reference,
        )

    @Slot(object)
    def _revive_db_reference(self, url):
        """Colors the db reference back to black.

        Args:
            url (URL): SqlAlchemy URL
        """
        url_text = remove_credentials_from_url(str(url))
        for row in range(self._db_ref_root.rowCount()):
            item = self._db_ref_root.child(row)
            if url_text == item.text():
                self._mark_as_found(item)
                url_dict = item.data(_Role.DB_URL_REFERENCE)
                if url_dict["dialect"] == "sqlite":
                    self.file_system_watcher.add_persistent_file_path(url_dict["database"])
                break
        else:
            return
        self._check_notifications()
        self._resources_to_successors_changed()

    @Slot(QModelIndex)
    def open_reference(self, index):
        """Open reference in default program."""
        if not index:
            return
        if not index.isValid():
            logging.error("Index not valid")
            return
        parent_item = self.reference_model.itemFromIndex(index.parent())
        if parent_item is not self._file_ref_root:
            return
        reference = self.file_references[index.row()]
        url = "file:///" + reference
        res = open_url(url)
        if not res:
            self._logger.msg_error.emit(f"Failed to open reference:<b>{reference}</b>")

    @Slot(QModelIndex)
    def open_data_file(self, index):
        """Open data file in default program."""
        if not index:
            return
        if not index.isValid():
            logging.error("Index not valid")
            return
        data_file = index.data(_Role.DATA_FILE_PATH)
        url = "file:///" + data_file
        # noinspection PyTypeChecker, PyCallByClass, PyArgumentList
        res = open_url(url)
        if not res:
            self._logger.msg_error.emit(f"Opening file <b>{data_file}</b> failed")

    def make_new_file(self):
        """Create a new blank file to this Data Connections data directory."""
        msg = "File name"
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QInputDialog.getText(
            self._toolbox, "Create new file", msg, flags=Qt.WindowTitleHint | Qt.WindowCloseButtonHint
        )
        file_name = answer[0]
        if not file_name.strip():
            return
        # Check that file name has no invalid chars
        if any(True for x in file_name if x in INVALID_FILENAME_CHARS):
            msg = f"File name <b>{file_name}</b> contains invalid characters."
            self._logger.information_box.emit("Creating file failed", msg)
            return
        file_path = os.path.join(self.data_dir, file_name)
        if os.path.exists(file_path):
            msg = f"File <b>{file_name}</b> already exists."
            self._logger.information_box.emit("Creating file failed", msg)
            return
        try:
            with open(file_path, "w"):
                self._logger.msg.emit(f"File <b>{file_name}</b> created to Data Connection <b>{self.name}</b>")
        except OSError:
            msg = "Please check directory permissions."
            self._logger.information_box.emit("Creating file failed", msg)

    @Slot()
    def remove_files(self):
        """Remove selected files from data directory."""
        indexes = self._properties_ui.treeView_dc_data.selectedIndexes()
        if not indexes:  # Nothing selected
            self._logger.msg.emit("Please select files to remove")
            return
        file_list = [index.data() for index in indexes]
        files = "\n".join(file_list)
        msg = "The following files will be removed permanently from the project\n\n" f"{files}\n\n" "Are you sure?"
        title = f"Remove {len(file_list)} File(s)"
        message_box = QMessageBox(
            QMessageBox.Icon.Question,
            title,
            msg,
            QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel,
            parent=self._toolbox,
        )
        message_box.button(QMessageBox.StandardButton.Ok).setText("Remove Files")
        answer = message_box.exec()
        if answer == QMessageBox.StandardButton.Cancel:
            return
        self.delete_files_from_project(file_list)

    def delete_files_from_project(self, file_names):
        """Deletes given files from item's data directory.

        Args:
            file_names (Iterable of str): files to delete
        """
        for filename in file_names:
            path_to_remove = os.path.join(self.data_dir, filename)
            try:
                os.remove(path_to_remove)
                self._logger.msg.emit(f"File <b>{path_to_remove}</b> removed")
            except OSError:
                self._logger.msg_error.emit(f"Removing file {path_to_remove} failed.\nCheck permissions.")

    def populate_reference_list(self, db_references):
        """List references in QTreeView.

        Args:
            db_references (list of dict): database URLs
        """
        self.reference_model.clear()
        self.reference_model.setHorizontalHeaderItem(0, QStandardItem("References"))  # Add header
        self._file_ref_root.removeRows(0, self._file_ref_root.rowCount())
        self.reference_model.appendRow(self._file_ref_root)
        self._db_ref_root.removeRows(0, self._db_ref_root.rowCount())
        self.reference_model.appendRow(self._db_ref_root)
        self._append_file_references_to_model(*self.file_references)
        self._append_db_references_to_model(*db_references)

    def _append_file_references_to_model(self, *paths):
        non_existent_paths = []
        for path in paths:
            item = QStandardItem(path)
            item.setFlags(~Qt.ItemIsEditable)
            if not os.path.exists(path):
                non_existent_paths.append(path)
                self._mark_as_missing(item)
            self._file_ref_root.appendRow(item)
        if non_existent_paths:
            msg = f"<b>{self.name}:</b> Could not find file references:"
            for path in non_existent_paths:
                msg += f"<br><b>{os.path.basename(path)}</b>"
            self._logger.msg_error.emit(msg)

    @staticmethod
    def _mark_as_missing(item):
        """Modifies given model item to appear as missing reference.

        Args:
            item (QStandardItem): item to modify
        """
        item.setData("The file is missing.", Qt.ItemDataRole.ToolTipRole)
        item.setData(_MISSING_ITEM_FOREGROUND, Qt.ItemDataRole.ForegroundRole)
        item.setData(True, _Role.MISSING)

    @staticmethod
    def _mark_as_found(item):
        """Modifies given model item to appear as existing reference.

        Args:
            item (QStandardItem): item to modify
        """
        item.setData(None, Qt.ItemDataRole.ToolTipRole)
        item.setData(None, Qt.ItemDataRole.ForegroundRole)
        item.setData(False, _Role.MISSING)

    def _append_db_references_to_model(self, *urls):
        """Appends given database URLs to the model.

        Args:
            *urls: dict-style URLs to add
        """
        for url in urls:
            item = QStandardItem(convert_url_to_safe_string(url))
            item.setData(url, _Role.DB_URL_REFERENCE)
            item.setFlags(~Qt.ItemIsEditable)
            self._db_ref_root.appendRow(item)

    def _watch_sqlite_file(self, *urls):
        """Adds sqlite files to file system watcher's watched paths.

        Args:
            *urls: dict-style URLs to watch
        """
        for url in urls:
            if url["dialect"] == "sqlite":
                path = url["database"]
                if os.path.exists(path):
                    self.file_system_watcher.add_persistent_file_path(path)

    def _unwatch_sqlite_file(self, *urls):
        """Removes sqlite files from file system watcher's watched paths.

        Args:
            *urls: dict-style URLs to watch

        Returns:
            list of str: list of removed paths
        """
        paths = []
        for url in urls:
            if url["dialect"] == "sqlite":
                path = url["database"]
                if os.path.exists(path):
                    paths.append(path)
        return self.file_system_watcher.remove_persistent_file_paths(paths)

    def populate_data_list(self):
        """List project internal data (files) in QTreeView."""
        self.data_model.clear()
        self.data_model.setHorizontalHeaderItem(0, QStandardItem("Data"))  # Add header
        self._append_data_files_to_model(*self.data_files())

    def _append_data_files_to_model(self, *paths):
        for path in paths:
            item = QStandardItem(os.path.basename(path))
            item.setFlags(~Qt.ItemIsEditable)
            icon = QFileIconProvider().icon(QFileInfo(path))
            item.setData(icon, Qt.ItemDataRole.DecorationRole)
            item.setData(path, _Role.DATA_FILE_PATH)
            self.data_model.appendRow(item)

    def resources_for_direct_successors(self):
        """see base class"""
        data_files = [os.path.join(self.data_dir, f) for f in self.data_files()]
        resources = scan_for_resources(
            self, self.file_references + data_files, list(self.db_reference_iter()), self._project.project_dir
        )
        return resources

    def _check_notifications(self):
        """Sets or clears the exclamation mark icon."""
        self.clear_notifications()
        if not self.file_references and not self.has_db_references() and not self.data_files():
            self.add_notification(
                "This Data Connection does not have any references or data. "
                "Add some in the Data Connection Properties panel."
            )
        missing_file_references = [ref for ref in self.file_references if not os.path.exists(ref)]
        if missing_file_references:
            self.add_notification("Cannot find some file references. Please, check that the files exist.")

    def item_dict(self):
        """Returns a dictionary corresponding to this item."""
        d = super().item_dict()
        d["file_references"] = [serialize_path(ref, self._project.project_dir) for ref in self.file_references]
        db_references = []
        db_credentials = {}
        for url in self.db_reference_iter():
            serialized_url = dict(url)
            username = serialized_url.pop("username")
            password = serialized_url.pop("password")
            if username:
                db_credentials[convert_url_to_safe_string(serialized_url)] = username, password
            if serialized_url["dialect"] == "sqlite":
                serialized_url["database"] = serialize_path(serialized_url["database"], self._project.project_dir)
            db_references.append(serialized_url)
        d["db_references"] = db_references
        d["db_credentials"] = db_credentials
        return d

    @staticmethod
    def item_dict_local_entries():
        """See base class."""
        return [("db_credentials",)]

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        # FIXME: Do we want to convert references to file_references via upgrade?
        file_references = item_dict.get("file_references", []) or item_dict.get("references", [])
        file_references = [deserialize_path(r, project.project_dir) for r in file_references]
        db_references = restore_database_references(
            item_dict.get("db_references", []), item_dict.get("db_credentials", {}), project.project_dir
        )
        return DataConnection(name, description, x, y, toolbox, project, file_references, db_references)

    def rename(self, new_name, rename_data_dir_message):
        """See base class."""
        old_data_dir = self.data_dir
        self.file_system_watcher.remove_persistent_dir_path(old_data_dir)
        if not super().rename(new_name, rename_data_dir_message):
            self.file_system_watcher.add_persistent_dir_path(old_data_dir)
            return False
        self.file_system_watcher.add_persistent_dir_path(self.data_dir)
        self.populate_data_list()
        return True

    def tear_down(self):
        """Tears down this item. Called by toolbox just before closing."""
        super().tear_down()
        self.file_system_watcher.tear_down()
        self._database_validator.wait_for_finish()

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Tool":
            self._logger.msg.emit(
                f"Link established. Tool <b>{source_item.name}</b> output files will be "
                f"passed as references to item <b>{self.name}</b> after execution."
            )
        elif source_item.item_type() in ["Data Store", "Importer"]:
            # Does this type of link do anything?
            self._logger.msg.emit("Link established")
        else:
            super().notify_destination(source_item)

    def temporary_paths_in_item_directory(self) -> list[pathlib.Path]:
        return [path for path in pathlib.Path(self.data_dir).iterdir() if not path.is_file()]



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains the source data table model."""
from PySide6.QtCore import QModelIndex, Qt, Signal, Slot
from spinedb_api import ParameterValueFormatError
from spinedb_api.import_mapping.type_conversion import ConvertSpec
from spinedb_api.mapping import Position
from spinetoolbox.mvcmodels.minimal_table_model import MinimalTableModel
from ..mapping_colors import ERROR_COLOR
from .mappings_model import Role


class SourceDataTableModel(MinimalTableModel):
    """A model for import mapping specification.

    Highlights columns, rows, and so on, depending on Mapping specification.
    """

    _FETCH_CHUNK_SIZE = 100
    more_data_needed = Signal(int, int)
    column_types_updated = Signal()
    row_types_updated = Signal()
    mapping_data_changed = Signal()
    polish_mapping_requested = Signal()
    """Emitted when mappings in the mapping list need polish."""
    about_to_undo = Signal(str)
    """Emitted when an undo/redo command is going to be executed."""

    def __init__(self, parent=None):
        """
        Args:
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._unfetched_data = []
        self._mapping_list_index = QModelIndex()
        self._column_types = {}
        self._row_types = {}
        self._column_type_errors = {}
        self._row_type_errors = {}
        self._converted_data = {}
        self._infinite = False
        self._infinite_extent = 0
        self._fetching = False

    def flags(self, index):
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable & ~Qt.ItemFlag.ItemIsEditable

    def clear(self, infinite=False):
        super().clear()
        self._column_type_errors = {}
        self._row_type_errors = {}
        self._column_types = {}
        self._row_types = {}
        self._converted_data = {}
        self._infinite = infinite
        self._infinite_extent = 0
        self._fetching = False

    def set_horizontal_header_labels(self, labels):
        super().set_horizontal_header_labels(labels)
        self.polish_mapping_requested.emit()

    def canFetchMore(self, parent):
        return not self._fetching

    def fetchMore(self, parent):
        if self._infinite:
            self.beginResetModel()
            self._infinite_extent += self._FETCH_CHUNK_SIZE // 10
            self.endResetModel()
            return
        self._fetching = True
        self.more_data_needed.emit(len(self._main_data) + self._FETCH_CHUNK_SIZE, len(self._main_data))

    def append_rows(self, data):
        self._fetching = False
        if not data:
            return
        self.beginInsertRows(QModelIndex(), len(self._main_data), len(self._main_data) + len(data) - 1)
        self._main_data += data
        self.endInsertRows()

    def rowCount(self, parent=QModelIndex()):
        if self._infinite:
            return self._infinite_extent
        return super().rowCount(parent)

    def columnCount(self, parent=QModelIndex()):
        if self._infinite:
            return self._infinite_extent
        return super().columnCount(parent)

    def set_mapping_list_index(self, index):
        """Set index to mappings model.

        Args:
            index (QModelIndex): index
        """
        self._mapping_list_index = index
        self.handle_mapping_data_changed()

    def validate(self, section, orientation=Qt.Orientation.Horizontal):
        converter = self.get_type(section, orientation)
        if converter is None:
            return
        if orientation == Qt.Orientation.Horizontal:
            for row in range(self.rowCount()):
                self._column_type_errors.pop((row, section), None)
                data = self.index(row, section).data(Qt.ItemDataRole.EditRole)
                if isinstance(data, str) and not data:
                    data = None
                if data is not None:
                    try:
                        self._converted_data[row, section] = converter(data)
                    except (ValueError, TypeError, ParameterValueFormatError) as e:
                        self._converted_data.pop((row, section), None)
                        self._column_type_errors[row, section] = e
            top_left = self.index(0, section)
            bottom_right = self.index(self.rowCount() - 1, section)
        else:
            for column in range(self.columnCount()):
                self._row_type_errors.pop((section, column), None)
                data = self.index(section, column).data(Qt.ItemDataRole.EditRole)
                if isinstance(data, str) and not data:
                    data = None
                if data is not None:
                    try:
                        self._converted_data[section, column] = converter(data)
                    except (ValueError, ParameterValueFormatError) as e:
                        self._converted_data.pop((section, column), None)
                        self._row_type_errors[section, column] = e
            top_left = self.index(section, 0)
            bottom_right = self.index(section, self.columnCount() - 1)
        self.dataChanged.emit(top_left, bottom_right)

    def get_type(self, section, orientation=Qt.Orientation.Horizontal):
        if orientation == Qt.Orientation.Horizontal:
            return self._column_types.get(section, None)
        return self._row_types.get(section, None)

    def get_types(self, orientation=Qt.Orientation.Horizontal):
        if orientation == Qt.Orientation.Horizontal:
            return self._column_types
        return self._row_types

    @Slot(int, object, object)
    def set_type(self, section, section_type, orientation=Qt.Orientation.Horizontal):
        if orientation == Qt.Orientation.Horizontal:
            count = self.columnCount()
            emit_signal = self.column_types_updated
            type_dict = self._column_types
        else:
            count = self.rowCount()
            emit_signal = self.row_types_updated
            type_dict = self._row_types
        if not isinstance(section_type, ConvertSpec):
            raise TypeError(
                f"section_type must be a instance of ConvertSpec, instead got {type(section_type).__name__}"
            )
        if section < 0 or section >= count:
            return
        type_dict[section] = section_type
        emit_signal.emit()
        self.validate(section, orientation)

    def set_types(self, sections, section_type, orientation):
        type_dict = self._column_types if orientation == Qt.Orientation.Horizontal else self._row_types
        for section in sections:
            type_dict[section] = section_type
            self.validate(section, orientation)
        if orientation == Qt.Orientation.Horizontal:
            self.column_types_updated.emit()
        else:
            self.row_types_updated.emit()

    @Slot(object, object)
    def set_all_column_types(self, excluded_columns, column_type):
        for column in range(self.columnCount()):
            if column not in excluded_columns:
                self._column_types[column] = column_type
        self.column_types_updated.emit()

    @Slot()
    def handle_mapping_data_changed(self):
        self.update_colors()
        self.mapping_data_changed.emit()

    def update_colors(self):
        top_left = self.index(0, 0)
        bottom_right = self.index(self.rowCount() - 1, self.columnCount() - 1)
        self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.BackgroundRole])
        self.headerDataChanged.emit(Qt.Orientation.Horizontal, 0, self.columnCount() - 1)

    def data_error(self, error, index, role=Qt.ItemDataRole.DisplayRole, orientation=Qt.Orientation.Horizontal):
        if role == Qt.ItemDataRole.ToolTipRole:
            type_display_name = self.get_type(index.column(), orientation).DISPLAY_NAME
            value = self._main_data[index.row()][index.column()]
            return f'<p>Could not parse value: "{value}" as a {type_display_name}: {error}</p>'
        if role == Qt.ItemDataRole.BackgroundRole:
            return ERROR_COLOR

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role in (Qt.ItemDataRole.ToolTipRole, Qt.ItemDataRole.BackgroundRole):
            flattened_mappings = None
            if self._mapping_list_index.isValid():
                flattened_mappings = self._mapping_list_index.data(Role.FLATTENED_MAPPINGS)
                last_pivot_row = flattened_mappings.root_mapping.last_pivot_row()
                read_start_row = flattened_mappings.root_mapping.read_start_row
            else:
                last_pivot_row = -1
                read_start_row = 0
            row = index.row()
            column = index.column()
            if index.row() > max(last_pivot_row, read_start_row - 1):
                error = self._column_type_errors.get((row, column))
                if error is not None:
                    return self.data_error(error, index, role, orientation=Qt.Orientation.Horizontal)

            if row <= last_pivot_row and column not in self._non_pivoted_and_skipped_columns(flattened_mappings):
                error = self._row_type_errors.get((row, column))
                if error is not None:
                    return self.data_error(error, index, role, orientation=Qt.Orientation.Vertical)

        if role == Qt.ItemDataRole.BackgroundRole and self._mapping_list_index.isValid():
            return self.data_color(index)
        if role == Qt.ItemDataRole.DisplayRole:
            converted_data = self._converted_data.get((index.row(), index.column()))
            if converted_data is not None:
                return str(converted_data)
        if self._infinite and role == Qt.ItemDataRole.DisplayRole:
            return f"item_{index.row() + 1}_{index.column() + 1}"
        return super().data(index, role)

    def _non_pivoted_and_skipped_columns(self, flattened_mappings=None):
        """Returns a list of non-pivoted and skipped columns.

        Args:
            flattened_mappings (FlattenedMappings, optional): flattened import mappings

        Returns:
            set: non-pivoted and skipped columns
        """
        if flattened_mappings is None:
            flattened_mappings = self._mapping_list_index.data(Role.FLATTENED_MAPPINGS)
        root_mapping = flattened_mappings.root_mapping
        return set(root_mapping.non_pivoted_columns()) | set(root_mapping.skip_columns)

    def data_color(self, index):
        """
        Returns background color for index depending on mapping.

        Arguments:
            index (QModelIndex): index

        Returns:
            QColor: color of index
        """
        row = index.row()
        flattened_mappings = self._mapping_list_index.data(Role.FLATTENED_MAPPINGS)
        if row < flattened_mappings.root_mapping.read_start_row:
            return None
        root_mapping = flattened_mappings.root_mapping
        row -= root_mapping.read_start_row
        column = index.column()
        last_row = root_mapping.last_pivot_row()
        non_pivoted_and_skipped_columns = self._non_pivoted_and_skipped_columns(flattened_mappings)
        if self.index_below_last_pivot_row(
            row, column, last_row, root_mapping.is_pivoted(), non_pivoted_and_skipped_columns
        ):
            return flattened_mappings.display_colors[-1]
        for k in range(len(flattened_mappings.display_names)):
            component_mapping = flattened_mappings.component_at(k)
            if self.index_in_mapping(component_mapping, row, column, last_row, non_pivoted_and_skipped_columns):
                return flattened_mappings.display_colors[k]
        return None

    @staticmethod
    def index_below_last_pivot_row(row, column, last_row, is_pivoted, non_pivoted_and_skipped_columns):
        """Checks if given index is outside pivot.

        Args:
            row (int): index row
            column (int): index column
            last_row (int): last non-pivoted row index
            is_pivoted (bool): whether mappings are pivoted
            non_pivoted_and_skipped_columns (set of int): set of column indexes

        Returns:
            bool: True if index is below the pivot, False otherwise
        """
        if not is_pivoted:
            return False
        return row > last_row and column not in non_pivoted_and_skipped_columns

    @staticmethod
    def index_in_mapping(mapping, row, column, last_row, non_pivoted_and_skipped_columns):
        """
        Checks if index is in mapping

        Args:
            mapping (ImportMapping): mapping component
            row (int): index row
            column (int): index column
            last_row (int): last non-pivoted row index
            non_pivoted_and_skipped_columns (set of int): set of column indexes

        Returns:
            bool: True if mapping is in index
        """
        if not isinstance(mapping.position, int):
            return False
        if mapping.position < 0:
            if column in non_pivoted_and_skipped_columns:
                return False
            return row == -(mapping.position + 1)
        if row <= last_row:
            return False
        return column == mapping.position

    def headerData(self, section, orientation=Qt.Orientation.Horizontal, role=Qt.ItemDataRole.DisplayRole):
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.BackgroundRole
            and self._mapping_list_index.isValid()
        ):
            flattened_mappings = self._mapping_list_index.data(Role.FLATTENED_MAPPINGS)
            for k in range(len(flattened_mappings.display_names)):
                component_mapping = flattened_mappings.component_at(k)
                if self.section_in_mapping(component_mapping, section):
                    return flattened_mappings.display_colors[k]
        if self._infinite and orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return f"header_{section + 1}"
        return super().headerData(section, orientation, role)

    def section_in_mapping(self, mapping, section):
        if mapping.position == Position.header and mapping.value is None:
            # Mapping from all headers
            return True
        return mapping.position == Position.header and mapping.value in (section, self.headerData(section))



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains controller that manages value transformations editor."""
from PySide6.QtCore import QModelIndex, QObject, Slot
from PySide6.QtWidgets import QFormLayout, QLineEdit
from ..commands import AppendInstruction, ChangeInstructionParameter, ChangeOperation, RemoveInstruction
from ..mvcmodels.value_transformations_table_model import TransformationsTableColumn, TransformationsTableRole

_DISPLAY_TEXT_TO_OPERATION = {
    "multiply": "multiply",
    "invert": "invert,",
    "negate": "negate",
    "generate index": "generate_index",
}

_OPERATION_TO_DISPLAY_TEXT = {op: label for label, op in _DISPLAY_TEXT_TO_OPERATION.items()}


class InstructionsEditor(QObject):
    def __init__(self, ui, transformations_table_model, undo_stack, parent):
        """
        Args:
            ui (Ui_Form): ValueTransformingWidget's interface
            transformations_table_model (ValueTransformationTableModel): transforms table model
            undo_stack (QUndoStack): undo stack
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._transformations_table_model = transformations_table_model
        self._transformations_table_index = QModelIndex()
        self._undo_stack = undo_stack
        self._ui = ui
        for operation_label in _DISPLAY_TEXT_TO_OPERATION:
            self._ui.operation_combo_box.addItem(operation_label)
        self._ui.add_instruction_button.clicked.connect(self._add_instruction)
        self._ui.remove_instruction_button.clicked.connect(self._ui.remove_instruction_action.trigger)
        self._ui.remove_instruction_action.triggered.connect(self._remove_instruction)
        self._ui.instructions_list_view.addAction(self._ui.remove_instruction_action)
        self._ui.instructions_list_view.currentRowChanged.connect(self._update_operation_editor)
        self._ui.operation_combo_box.currentTextChanged.connect(self._change_operation)
        self._ui.transformations_table_view.selectionModel().currentChanged.connect(self._transformation_changed)

    @Slot(bool)
    def _add_instruction(self, checked):
        """Pushes an append instruction command to undo stack.

        Args:
            checked (bool): unused
        """
        operation = _DISPLAY_TEXT_TO_OPERATION[self._ui.operation_combo_box.itemText(0)]
        instruction = {"operation": operation}
        if operation == "multiply":
            instruction["rhs"] = 1.0
        elif operation == "generate_index":
            instruction["expression"] = ""
        self._undo_stack.push(AppendInstruction(self, instruction))

    def append_instruction(self, instruction):
        """Appends an instruction to instructions list.

        Args:
            instruction (dict): instruction to append
        """
        row = self._ui.instructions_list_view.count()
        self.insert_instruction(row, instruction)

    def insert_instruction(self, i, instruction):
        """Inserts an instruction to instructions list.

        Args:
            i (int): insertion point
            instruction (dict): instruction to append
        """
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instructions.insert(i, instruction)
        self._transformations_table_model.setData(
            self._transformations_table_index, instructions, TransformationsTableRole.INSTRUCTIONS
        )
        self._ui.instructions_list_view.insertItem(i, instruction["operation"])
        self._ui.instructions_list_view.setCurrentRow(i)
        self._ui.operation_combo_box.setEnabled(True)

    @Slot(bool)
    def _remove_instruction(self, checked):
        """Pushes a remove instruction command to undo stack.

        Args:
            checked (bool): unused
        """
        row = self._ui.instructions_list_view.currentRow()
        if row < 0:
            return
        self._undo_stack.push(RemoveInstruction(self, row))

    def pop_instruction(self, i):
        """Pops last instruction from instructions list.

        Args:
            i (int): instruction index; if negative, pop from back
        """
        row = i if i >= 0 else self._ui.instructions_list_view.count() + i
        self._ui.instructions_list_view.takeItem(row)
        if self._ui.instructions_list_view.count() == 0:
            self._ui.operation_combo_box.setEnabled(False)
            self._remove_operation_widgets()
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instructions.pop(i)
        self._transformations_table_model.setData(
            self._transformations_table_index, instructions, TransformationsTableRole.INSTRUCTIONS
        )

    def _load_instructions(self, row):
        """Updates widgets.

        Args:
            row (int): row index to ValueTransformationsTableModel
        """
        if row == -1:
            self._transformations_table_index = QModelIndex()
            self._ui.instructions_list_view.clear()
            return
        self._transformations_table_index = self._transformations_table_model.index(
            row, TransformationsTableColumn.INSTRUCTIONS
        )
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        self._ui.instructions_list_view.clear()
        self._ui.instructions_list_view.addItems([t["operation"] for t in instructions])
        self._enable_widgets()

    @Slot(str)
    def _change_operation(self, operation_label):
        """Pushes a change operation command to undo stack.

        Args:
            operation_label (str): operation's display text
        """
        row = self._ui.instructions_list_view.currentRow()
        self._undo_stack.push(ChangeOperation(self, row, _DISPLAY_TEXT_TO_OPERATION[operation_label]))

    def set_operation(self, i, operation):
        """Sets an operation for instruction.

        Args:
            i (int), instruction index
            operation (str): operation
        """
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instruction = {"operation": operation}
        if operation == "multiply":
            instruction["rhs"] = 1.0
        elif operation == "generate_index":
            instruction["expression"] = ""
        instructions[i] = instruction
        self._transformations_table_model.setData(
            self._transformations_table_index, instructions, TransformationsTableRole.INSTRUCTIONS
        )
        self._ui.instructions_list_view.item(i).setText(operation)
        if i == self._ui.instructions_list_view.currentRow():
            self._update_operation_editor(i)

    def instruction(self, i):
        """Returns instruction from index.

        Args:
            i (int): instruction index

        Returns:
            dict: instruction
        """
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        return instructions[i]

    def set_instruction(self, i, instruction):
        """Sets instruction.

        Args:
            i (int): instruction index
            instruction (dict): instruction
        """
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instructions[i] = instruction
        self._transformations_table_model.setData(
            self._transformations_table_index, instructions, TransformationsTableRole.INSTRUCTIONS
        )
        self._ui.instructions_list_view.item(i).setText(instruction["operation"])
        if i == self._ui.instructions_list_view.currentRow():
            self._update_operation_editor(i)

    def _enable_widgets(self):
        """Enables or disables widgets that deal with value operations."""
        enabled = self._ui.transformations_table_view.selectionModel().currentIndex().isValid()
        self._ui.add_instruction_button.setEnabled(enabled)
        self._ui.remove_instruction_button.setEnabled(enabled)
        self._ui.instructions_list_view.setEnabled(enabled)
        enabled = enabled and self._ui.instructions_list_view.currentRow() >= 0
        self._ui.operation_combo_box.setEnabled(enabled)

    @Slot(int)
    def _update_operation_editor(self, row):
        """Updates operation editor widgets.

        Args:
            row (int): row index to operations list view
        """
        if row == -1:
            self._enable_widgets()
            return
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instruction = instructions[row]
        operation = instruction["operation"]
        self._ui.operation_combo_box.currentTextChanged.disconnect(self._change_operation)
        self._ui.operation_combo_box.setCurrentText(_OPERATION_TO_DISPLAY_TEXT[operation])
        self._ui.operation_combo_box.currentTextChanged.connect(self._change_operation)
        self._make_operation_widgets(operation)
        self._update_operation_widgets()
        self._enable_widgets()

    def _make_operation_widgets(self, operation):
        """Creates operation-specific widgets and adds them to instruction options layout.

        Args:
            operation (str): operation
        """
        self._remove_operation_widgets()
        if operation == "multiply":
            edit = QLineEdit()
            edit.editingFinished.connect(self._change_multiply_multiplier)
            self._ui.instruction_options_layout.addRow("Multiplier:", edit)
        elif operation == "generate_index":
            edit = QLineEdit()
            edit.setPlaceholderText("Enter Python expression...")
            edit.setToolTip("A Python expression, e.g. f'T{i:03}' for T001, T002,...")
            edit.editingFinished.connect(self._change_index_generator_expression)
            self._ui.instruction_options_layout.addRow("Expression:", edit)

    def _remove_operation_widgets(self):
        """Removes operation-specific widgets from instruction option layout."""
        while self._ui.instruction_options_layout.rowCount() > 1:
            self._ui.instruction_options_layout.removeRow(1)

    @Slot(QModelIndex, QModelIndex)
    def _transformation_changed(self, current, previous):
        """Loads new instructions to the editor.

        Args:
            current (QModelIndex): new index in parameters table
            previous (QModelIndex): previous index in parameters table
        """
        sorted_model = self._ui.transformations_table_view.model()
        current = sorted_model.mapToSource(current)
        if current.row() == self._transformations_table_index.row():
            return
        self._remove_operation_widgets()
        self._load_instructions(current.row())

    def _update_operation_widgets(self):
        """Updates operation-specific widgets."""
        row = self._ui.instructions_list_view.currentRow()
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instruction = instructions[row]
        operation = instruction["operation"]
        if operation == "multiply":
            edit = self._ui.instruction_options_layout.itemAt(1, QFormLayout.FieldRole).widget()
            edit.setText(str(instruction["rhs"]))
        elif operation == "generate_index":
            edit = self._ui.instruction_options_layout.itemAt(1, QFormLayout.FieldRole).widget()
            edit.setText(str(instruction["expression"]))

    @Slot()
    def _change_multiply_multiplier(self):
        """Pushes a command to change multiplier coefficient to undo stack."""
        row = self._ui.instructions_list_view.currentRow()
        edit = self._ui.instruction_options_layout.itemAt(1, QFormLayout.FieldRole).widget()
        try:
            multiplier = float(edit.text())
        except ValueError:
            return
        previous_multiplier = self.instruction(row)["rhs"]
        self._undo_stack.push(ChangeInstructionParameter(row, multiplier, previous_multiplier, self._set_multiplier))

    def _set_multiplier(self, i, multiplier):
        """Sets multiply operation's multiplier.

        Args:
            i (int): instruction index
            multiplier (float): new multiplier
        """
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instructions[i]["rhs"] = multiplier
        self._transformations_table_model.setData(
            self._transformations_table_index, instructions, TransformationsTableRole.INSTRUCTIONS
        )
        edit = self._ui.instruction_options_layout.itemAt(1, QFormLayout.FieldRole).widget()
        if i == self._ui.instructions_list_view.currentRow() and str(multiplier) != edit.text():
            edit.setText(str(multiplier))

    @Slot()
    def _change_index_generator_expression(self):
        """Pushes a command that changes index generator expression to undo stack."""
        row = self._ui.instructions_list_view.currentRow()
        edit = self._ui.instruction_options_layout.itemAt(1, QFormLayout.FieldRole).widget()
        expression = edit.text()
        try:
            eval(expression, {}, {"i": 1})
        except (AttributeError, NameError, SyntaxError, ValueError):
            self._ui.statusbar.showMessage("Invalid expression for index generator.", timeout=5000)
            if not edit.hasFocus():
                edit.setText(self.instruction(row)["expression"])
            return
        self._ui.statusbar.clearMessage()
        previous_expression = self.instruction(row)["expression"]
        self._undo_stack.push(ChangeInstructionParameter(row, expression, previous_expression, self._set_expression))

    def _set_expression(self, i, expression):
        """Sets index generator expression.

        Args:
            i (int): instruction index
            expression (str): new expression
        """
        instructions = self._transformations_table_model.data(
            self._transformations_table_index, TransformationsTableRole.INSTRUCTIONS
        )
        instructions[i]["expression"] = expression
        self._transformations_table_model.setData(
            self._transformations_table_index, instructions, TransformationsTableRole.INSTRUCTIONS
        )
        edit = self._ui.instruction_options_layout.itemAt(1, QFormLayout.FieldRole).widget()
        if i == self._ui.instructions_list_view.currentRow() and str(expression) != edit.text():
            edit.setText(expression)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Importer project item class."""
from operator import itemgetter
import os
from PySide6.QtCore import QModelIndex, Qt, Slot
from spinetoolbox.helpers import create_dir
from spinetoolbox.widgets.custom_menus import ItemSpecificationMenu
from ..commands import ChangeItemSelectionCommand, UpdateCancelOnErrorCommand, UpdateOnConflictCommand
from ..db_writer_item_base import DBWriterItemBase
from ..models import CheckableFileListModel
from .executable_item import ExecutableItem
from .item_info import ItemInfo


class Importer(DBWriterItemBase):
    def __init__(
        self,
        name,
        description,
        x,
        y,
        toolbox,
        project,
        specification_name="",
        cancel_on_error=True,
        on_conflict="merge",
        file_selection=None,
    ):
        """Importer class.

        Args:
            name (str): Project item name
            description (str): Project item description
            x (float): Initial icon scene X coordinate
            y (float): Initial icon scene Y coordinate
            toolbox (ToolboxUI): QMainWindow instance
            project (SpineToolboxProject): the project this item belongs to
            specification_name (str, optional): a spec name
            cancel_on_error (bool): if True the item's execution will stop on import error
            on_conflict (str): how to handle conflicts between parallel importers
            file_selection (dict, optional): a map from label to a bool indicating if the file item is checked
        """
        super().__init__(name, description, x, y, project)
        # Make logs subdirectory for this item
        self._toolbox = toolbox
        self.logs_dir = os.path.join(self.data_dir, "logs")
        try:
            create_dir(self.logs_dir)
        except OSError:
            self._logger.msg_error.emit(f"[OSError] Creating directory {self.logs_dir} failed. Check permissions.")
        # Variables for saving selections when item is (de)activated
        self._specification = self._project.get_specification(specification_name)
        if specification_name and not self._specification:
            self._logger.msg_error.emit(
                f"Importer <b>{self.name}</b> uses specification <b>{specification_name}</b> but it was not found"
            )
        self.cancel_on_error = cancel_on_error
        self.on_conflict = on_conflict
        self._file_model = CheckableFileListModel(header_label="Available resources")
        self._file_model.set_initial_state(file_selection if file_selection is not None else {})
        self._file_model.checked_state_changed.connect(self._push_file_selection_change_to_undo_stack)
        self._file_model.dataChanged.connect(self.check_notifs)

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.toolButton_edit_specification.clicked] = self.edit_specification
        s[self._properties_ui.treeView_files.doubleClicked] = self._handle_files_double_clicked
        s[self._properties_ui.comboBox_specification.textActivated] = self._change_specification
        s[self._properties_ui.cancel_on_error_checkBox.stateChanged] = self._handle_cancel_on_error_changed
        s[self._properties_ui.radioButton_on_conflict_merge.clicked] = self._update_on_conflict
        s[self._properties_ui.radioButton_on_conflict_keep.clicked] = self._update_on_conflict
        s[self._properties_ui.radioButton_on_conflict_replace.clicked] = self._update_on_conflict
        return s

    @Slot(str)
    def _change_specification(self, specification_name):
        """
        Pushes a set specification command to the undo stack.

        Args:
            specification_name (str): new specification name
        """
        specification = self._project.get_specification(specification_name)
        self.set_specification(specification)

    @Slot(int)
    def _handle_cancel_on_error_changed(self, _state):
        cancel_on_error = self._properties_ui.cancel_on_error_checkBox.isChecked()
        if self.cancel_on_error == cancel_on_error:
            return
        self._toolbox.undo_stack.push(UpdateCancelOnErrorCommand(self.name, cancel_on_error, self._project))

    def set_cancel_on_error(self, cancel_on_error):
        """Sets cancel on error setting.

        Args:
            cancel_on_error (bool): cancel on error flag
        """
        self.cancel_on_error = cancel_on_error
        if not self._active:
            return
        check_state = Qt.CheckState.Checked if self.cancel_on_error else Qt.CheckState.Unchecked
        self._properties_ui.cancel_on_error_checkBox.blockSignals(True)
        self._properties_ui.cancel_on_error_checkBox.setCheckState(check_state)
        self._properties_ui.cancel_on_error_checkBox.blockSignals(False)

    def _on_conflict(self):
        """Reads the on_conflict strategy from UI."""
        strategies = {
            self._properties_ui.radioButton_on_conflict_merge: "merge",
            self._properties_ui.radioButton_on_conflict_replace: "replace",
            self._properties_ui.radioButton_on_conflict_keep: "keep",
        }
        return strategies[next(button for button in strategies if button.isChecked())]

    def _set_on_conflict(self):
        """Sets the on_conflict strategy in the UI."""
        {
            "merge": self._properties_ui.radioButton_on_conflict_merge,
            "replace": self._properties_ui.radioButton_on_conflict_replace,
            "keep": self._properties_ui.radioButton_on_conflict_keep,
        }[self.on_conflict].setChecked(True)

    @Slot(bool)
    def _update_on_conflict(self, _checked):
        on_conflict = self._on_conflict()
        if self.on_conflict == on_conflict:
            return
        self._toolbox.undo_stack.push(UpdateOnConflictCommand(self.name, on_conflict, self._project))

    def set_on_conflict(self, on_conflict):
        """Sets on_conflict setting.

        Args:
            on_conflict (str): on_conflict
        """
        self.on_conflict = on_conflict
        if not self._active:
            return
        self._set_on_conflict()

    def restore_selections(self):
        """Restores selections into shared widgets when this project item is selected."""
        self._properties_ui.cancel_on_error_checkBox.setCheckState(
            Qt.CheckState.Checked if self.cancel_on_error else Qt.CheckState.Unchecked
        )
        self._set_on_conflict()
        self._properties_ui.treeView_files.setModel(self._file_model)
        self._update_ui()

    def _update_ui(self):
        if self._specification:
            self._properties_ui.comboBox_specification.setCurrentText(self._specification.name)
            spec_model_index = self._toolbox.specification_model.specification_index(self._specification.name)
            specification_options_popup_menu = ItemSpecificationMenu(self._toolbox, spec_model_index, self)
            self._properties_ui.toolButton_edit_specification.setMenu(specification_options_popup_menu)
        else:
            self._properties_ui.comboBox_specification.setCurrentIndex(-1)
            self._properties_ui.toolButton_edit_specification.setMenu(None)

    def save_selections(self):
        """Saves selections in shared widgets for this project item into instance variables."""
        self._properties_ui.treeView_files.setModel(None)

    def do_set_specification(self, specification):
        """see base class"""
        if not super().do_set_specification(specification):
            return False
        if self._active:
            self._update_ui()
        self._check_notifications()
        return True

    @Slot(bool)
    def edit_specification(self, checked=False):
        """Opens Import editor for the file selected in list view."""
        index = self._properties_ui.treeView_files.currentIndex()
        if not index.isValid() or not self._file_model.is_checked(index):
            index = self._file_model.index_with_file_path()
        self.open_import_editor(index)

    @Slot(QModelIndex)
    def _handle_files_double_clicked(self, index):
        """Opens Import editor for the double clicked index."""
        self.open_import_editor(index)

    def open_import_editor(self, index):
        """Opens Import editor for the given index.

        Args:
            index (QModelIndex): resource list index
        """
        source = None
        source_extras = None
        if index.isValid():
            resource = self._file_model.resource(index)
            if resource.type_ == "url":
                source = resource.url
                source_extras = resource.metadata
            else:
                if not resource.hasfilepath:
                    self._logger.msg_error.emit("File does not exist yet.")
                else:
                    if not os.path.exists(resource.path):
                        self._logger.msg_error.emit(f"Cannot find file '{source}'.")
                    else:
                        source = resource.path
        self._toolbox.show_specification_form(
            self.item_type(), self.specification(), self, source=source, source_extras=source_extras
        )

    @Slot(QModelIndex, bool)
    def _push_file_selection_change_to_undo_stack(self, index, checked):
        """Makes changes to file selection undoable.

        Args:
            index (QModelIndex): index to file model
            checked (bool): True if item was checked, False otherwise
        """
        self._toolbox.undo_stack.push(ChangeItemSelectionCommand(self.name, self._file_model, index, checked))

    def upstream_resources_updated(self, resources):
        self._file_model.update(resources)
        self._check_notifications()

    def replace_resources_from_upstream(self, old, new):
        """See base class."""
        for old_resource, new_resource in zip(old, new):
            self._file_model.replace(old_resource, new_resource)

    @Slot(QModelIndex, QModelIndex, list)
    def check_notifs(self, _tl_ind, _br_ind, _roles):
        """Checks notifications when the data in the resource model changes. E.g.
        when a resource is checked or unchecked."""
        self._check_notifications()

    def _check_notifications(self):
        self.clear_notifications()
        self._check_write_index()
        if not self.specification():
            self.add_notification(
                "This Importer does not have a specification. Set it in the Importer Properties Panel."
            )
            return
        duplicates = self._file_model.duplicate_paths()
        if duplicates:
            self.add_notification(f"Duplicate input files from upstream items:<br>{'<br>'.join(duplicates)}")
            return
        if self._file_model.rowCount() == 0:
            self.add_notification(
                "This Importer does not have any input data. "
                "Connect Data Connections to this Importer to use their data as input."
            )
            return
        any_selected = list()
        for row in range(self._file_model.rowCount()):
            label, selected = self._file_model.checked_data(self._file_model.index(row, 0))
            any_selected.append(selected)
        if not any(any_selected):
            self.add_notification(
                "No source files selected. Select <b>available resources</b> " "from this Importer's Properties."
            )

    def item_dict(self):
        """Returns a dictionary corresponding to this item."""
        d = super().item_dict()
        if not self.specification():
            d["specification"] = ""
        else:
            d["specification"] = self.specification().name
        d["cancel_on_error"] = self.cancel_on_error
        d["on_conflict"] = self.on_conflict
        selections = []
        for row in range(self._file_model.rowCount()):
            label, selected = self._file_model.checked_data(self._file_model.index(row, 0))
            selections.append([label, selected])
        d["file_selection"] = sorted(selections, key=itemgetter(0))
        return d

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class."""
        description, x, y = DBWriterItemBase.parse_item_dict(item_dict)
        specification_name = item_dict.get("specification", "")
        cancel_on_error = item_dict.get("cancel_on_error", False)
        on_conflict = item_dict.get("on_conflict", "merge")
        file_selection = dict(item_dict.get("file_selection", []))
        return Importer(
            name, description, x, y, toolbox, project, specification_name, cancel_on_error, on_conflict, file_selection
        )

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() in ("Data Connection", "Tool", "Exporter"):
            self._logger.msg.emit(
                "Link established. You can define mappings on data from "
                f"<b>{source_item.name}</b> using <b>{self.name}</b>."
            )
        elif source_item.item_type() == "Data Store":
            self._logger.msg.emit("Link established")
        else:
            super().notify_destination(source_item)

    @staticmethod
    def upgrade_v1_to_v2(item_name, item_dict):
        """Upgrades item's dictionary from v1 to v2.

        Changes:
        - if file_selection does not exist or if it
        is a list of booleans, reset file_selection
        to an empty list.

        Args:
            item_name (str): item's name
            item_dict (dict): Version 1 item dictionary

        Returns:
            dict: Version 2 Exporter dictionary
        """
        try:
            file_selection = item_dict["file_selection"]
        except KeyError:
            item_dict["file_selection"] = []
            return item_dict
        if len(file_selection) == 0:
            return item_dict
        if isinstance(file_selection[0], bool):
            item_dict["file_selection"] = []
        return item_dict

    @staticmethod
    def upgrade_v2_to_v3(item_name, item_dict, project_upgrader):
        """
        Upgrades item's dictionary from v2 to v3.

        1. Get rid of "mappings" in favor of "specification".
           The "mappings" are being turned into specifications by the project_upgrader
        2. Rename "mapping_selection" to "file_selection"

        Args:
            item_name (str): item's name
            item_dict (dict): Version 1 item dictionary
            project_upgrader (ProjectUpgrader)

        Returns:
            dict: Version 2 item dictionary
        """
        mappings = item_dict["mappings"]
        mapping = next(iter(mappings), None)
        if mapping is None:
            specification_name = ""
        else:
            label, _ = mapping
            specification_name = project_upgrader.make_unique_importer_specification_name(item_name, label, 0)
        new_item_dict = {k: v for k, v in item_dict.items() if k != "mappings"}
        new_item_dict["specification"] = specification_name
        new_item_dict["file_selection"] = item_dict.pop("mapping_selection")
        return new_item_dict


def _fix_csv_reader_settings(settings):
    """CSVReader saved the table names as the filepath, change that
    to 'csv' instead. This function will mutate the dictionary.

    Args:
        settings (dict): Mapping settings that should be updated
    """
    table_mappings = settings.get("table_mappings", {})
    k = list(table_mappings)
    if len(k) == 1 and k[0] != "csv":
        table_mappings["csv"] = table_mappings.pop(k[0])

    table_types = settings.get("table_types", {})
    k = list(table_types.keys())
    if len(k) == 1 and k[0] != "csv":
        table_types["csv"] = table_types.pop(k[0])

    table_row_types = settings.get("table_row_types", {})
    k = list(table_row_types.keys())
    if len(k) == 1 and k[0] != "csv":
        table_row_types["csv"] = table_row_types.pop(k[0])

    table_options = settings.get("table_options", {})
    k = list(table_options.keys())
    if len(k) == 1 and k[0] != "csv":
        table_options["csv"] = table_options.pop(k[0])

    selected_tables = settings.get("selected_tables", [])
    if len(selected_tables) == 1 and selected_tables[0] != "csv":
        selected_tables.pop(0)
        selected_tables.append("csv")



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Data transformer's undo commands."""
from PySide6.QtGui import QUndoCommand


class SetData(QUndoCommand):
    """Sets model's value."""

    def __init__(self, message, index, value, previous_value, role):
        """
        Args:
            message (str): undo message
            index (QModelIndex): model index
            value (Any): new value
            previous_value (Any): undo value
            role (int): set data role
        """
        super().__init__(message)
        self._index = index
        self._value = value
        self._previous_value = previous_value
        self._role = role

    def redo(self):
        success = self._index.model().setData(self._index, self._value, self._role)
        self.setObsolete(not success)

    def undo(self):
        self._index.model().setData(self._index, self._previous_value, self._role)


class InsertRow(QUndoCommand):
    """Inserts a row to value transformation model."""

    def __init__(self, message, model, row, data):
        """
        Args:
            message (str): undo message
            model (ValueTransformationsTableModel): model
            row (int): row index
            data (Sequence, optional): row data
        """
        super().__init__(message)
        self._model = model
        self._row = row
        self._data = data

    def redo(self):
        self._model.insertRow(self._row)
        if self._data is None:
            return
        for column, (element, role) in enumerate(zip(self._data, self._model.SET_DATA_ROLES)):
            index = self._model.index(self._row, column)
            self._model.setData(index, element, role)

    def undo(self):
        self._model.removeRow(self._row)


class RemoveRow(QUndoCommand):
    """Removes a row from value transformation model."""

    def __init__(self, message, model, row):
        """
        Args:
            message (str): undo message
            model (ValueTransformationsTableModel): model
            row (int): row index
        """
        super().__init__(message)
        self._model = model
        self._row = row
        self._previous_data = tuple(model.index(row, col).data(role) for col, role in enumerate(model.GET_DATA_ROLES))

    def redo(self):
        self._model.removeRow(self._row)

    def undo(self):
        self._model.insertRow(self._row)
        for column, (element, role) in enumerate(zip(self._previous_data, self._model.SET_DATA_ROLES)):
            index = self._model.index(self._row, column)
            self._model.setData(index, element, role)


class AppendInstruction(QUndoCommand):
    """Appends an instruction to value transformation."""

    def __init__(self, editor, instruction):
        """
        Args:
            editor (InstructionsEditor): instructions editor

            instruction (dict): instruction to append
        """
        super().__init__("add instruction")
        self._editor = editor
        self._instruction = instruction

    def redo(self):
        self._editor.append_instruction(self._instruction)

    def undo(self):
        self._editor.pop_instruction(-1)


class RemoveInstruction(QUndoCommand):
    """Removes an instruction from value transformation."""

    def __init__(self, editor, row):
        """
        Args:
            editor (InstructionsEditor): instructions editor
            row (int): instruction row index
        """
        super().__init__("remove instruction")
        self._editor = editor
        self._row = row
        self._instruction = self._editor.instruction(self._row)

    def redo(self):
        self._editor.pop_instruction(self._row)

    def undo(self):
        self._editor.insert_instruction(self._row, self._instruction)


class ChangeOperation(QUndoCommand):
    """Changes instruction's operation."""

    def __init__(self, editor, row, operation):
        """
        Args:
            editor (InstructionsEditor): instructions editor
            row (int): instruction row index
            operation (str): operation's name
        """
        super().__init__("change operation")
        self._editor = editor
        self._row = row
        self._operation = operation
        self._previous_instruction = self._editor.instruction(self._row)

    def redo(self):
        self._editor.set_operation(self._row, self._operation)

    def undo(self):
        self._editor.set_instruction(self._row, self._previous_instruction)


class ChangeInstructionParameter(QUndoCommand):
    """Changes multiply operations multiplier."""

    def __init__(self, row, parameter, previous_parameter, callback):
        """
        Args:
            row (int): instruction row index
            parameter (Any): parameter's new value
            previous_parameter (Any): parameter's old value
            callback (Callable): function to call to change the value
        """
        super().__init__("multiplier")
        self._row = row
        self._parameter = parameter
        self._previous_parameter = previous_parameter
        self._callback = callback
        if self._parameter == self._previous_parameter:
            self.setObsolete(True)

    def redo(self):
        self._callback(self._row, self._parameter)

    def undo(self):
        self._callback(self._row, self._previous_parameter)
